# validated: 2015-12-24 DS 6d854af shared/java/edu/wpi/first/wpilibj/livewindow/LiveWindow.java

from networktables import NetworkTable
from .command import Scheduler

import logging
logger = logging.getLogger(__name__)

__all__ = ["LiveWindow"]

class _LiveWindowComponent:
    """A LiveWindow component is a device (sensor or actuator) that should be
    added to the SmartDashboard in test mode. The components are cached until
    the first time the robot enters Test mode. This allows the components to
    be inserted, then renamed."""
    def __init__(self, subsystem, name, isSensor):
        self.subsystem = subsystem
        self.name = str(name)
        self.isSensor = isSensor

class LiveWindow:
    """The public interface for putting sensors and
    actuators on the LiveWindow."""

    sensors = set()
    #actuators = set()
    components = {}
    livewindowTable = None
    statusTable = None
    liveWindowEnabled = False
    firstTime = True
    
    @staticmethod
    def _reset():
        LiveWindow.sensors = set()
        LiveWindow.components = {}
        LiveWindow.livewindowTable = None
        LiveWindow.statusTable = None
        LiveWindow.liveWindowEnabled = False
        LiveWindow.firstTime = True

    @staticmethod
    def initializeLiveWindowComponents():
        """Initialize all the LiveWindow elements the first time we enter
        LiveWindow mode. By holding off creating the NetworkTable entries, it
        allows them to be redefined before the first time in LiveWindow mode.
        This allows default sensor and actuator values to be created that are
        replaced with the custom names from users calling addActuator and
        addSensor.
        """
        logger.info("Initializing the components first time")
        LiveWindow.livewindowTable = NetworkTable.getTable("LiveWindow")
        LiveWindow.statusTable = LiveWindow.livewindowTable.getSubTable("~STATUS~")
        for component, c in LiveWindow.components.items():
            logger.info("Initializing table for '%s' '%s'" % (c.subsystem, c.name))
            LiveWindow.livewindowTable.getSubTable(c.subsystem).putString("~TYPE~", "LW Subsystem")
            table = LiveWindow.livewindowTable.getSubTable(c.subsystem).getSubTable(c.name)
            table.putString("~TYPE~", component.getSmartDashboardType())
            table.putString("Name", c.name)
            table.putString("Subsystem", c.subsystem)
            component.initTable(table)
            if c.isSensor:
                LiveWindow.sensors.add(component)

    @staticmethod
    def setEnabled(enabled):
        """Set the enabled state of LiveWindow. If it's being enabled, turn
        off the scheduler and remove all the commands from the queue and
        enable all the components registered for LiveWindow. If it's being
        disabled, stop all the registered components and reenable the
        scheduler.

        TODO: add code to disable PID loops when enabling LiveWindow. The
        commands should reenable the PID loops themselves when they get
        rescheduled. This prevents arms from starting to move around, etc.
        after a period of adjusting them in LiveWindow mode.
        """
        if LiveWindow.liveWindowEnabled != enabled:
            if enabled:
                logger.info("Starting live window mode.")
                if LiveWindow.firstTime:
                    LiveWindow.initializeLiveWindowComponents()
                    LiveWindow.firstTime = False
                Scheduler.getInstance().disable()
                Scheduler.getInstance().removeAll()
                bad_components = []
                for component in LiveWindow.components.keys():
                    try:
                        component.startLiveWindowMode()
                    except Exception as e:
                        logger.error("Exception running startLiveWindowMode() on {}, removing from component list.".format(component))
                        logger.exception(e)
                        bad_components.append(component)
                for component in bad_components:
                    del(LiveWindow.components[component])
            else:
                logger.info("Stopping live window mode.")
                for component in LiveWindow.components.keys():
                    component.stopLiveWindowMode()
                Scheduler.getInstance().enable()
            LiveWindow.liveWindowEnabled = enabled
            LiveWindow.statusTable.putBoolean("LW Enabled", enabled)

    @staticmethod
    def run():
        """The run method is called repeatedly to keep the values refreshed
        on the screen in test mode.
        """
        LiveWindow.updateValues()

    @staticmethod
    def addSensor(subsystem, name, component):
        """Add a Sensor associated with the subsystem and with call it by the
        given name.

        :param subsystem: The subsystem this component is part of.
        :param name: The name of this component.
        :param component: A LiveWindowSendable component that represents a
            sensor.
        """
        LiveWindow.components[component] = \
                _LiveWindowComponent(subsystem, name, True)
        LiveWindow.sensors.add(component)

    @staticmethod
    def addActuator(subsystem, name, component):
        """Add an Actuator associated with the subsystem and with call it by
        the given name.

        :param subsystem: The subsystem this component is part of.
        :param name: The name of this component.
        :param component: A LiveWindowSendable component that represents a
            actuator.
        """
        LiveWindow.components[component] = \
                _LiveWindowComponent(subsystem, name, False)

    @staticmethod
    def updateValues():
        """Puts all sensor values on the live window."""
        # TODO: gross - needs to be sped up
        for lws in LiveWindow.sensors:
            lws.updateTable()
        # TODO: Add actuators?
        # TODO: Add better rate limiting.

    @staticmethod
    def addSensorChannel(moduleType, channel, component):
        """Add Sensor to LiveWindow. The components are shown with the type
        and channel like this: Gyro[0] for a gyro object connected to the
        first analog channel.

        :param moduleType: A string indicating the type of the module used in
            the naming (above)
        :param channel: The channel number the device is connected to
        :param component: A reference to the object being added
        """
        LiveWindow.addSensor("Ungrouped", "%s[%s]" % (moduleType, channel),
                             component)

    @staticmethod
    def addActuatorChannel(moduleType, channel, component):
        """Add Actuator to LiveWindow. The components are shown with the
        module type, slot and channel like this: Servo[0,2] for a servo
        object connected to the first digital module and PWM port 2.

        :param moduleType: A string that defines the module name in the label
            for the value
        :param channel: The channel number the device is plugged into
            (usually PWM)
        :param component: The reference to the object being added
        """
        LiveWindow.addActuator("Ungrouped", "%s[%s]" % (moduleType, channel),
                               component)

    @staticmethod
    def addActuatorModuleChannel(moduleType, moduleNumber, channel, component):
        """Add Actuator to LiveWindow. The components are shown with the
        module type, slot and channel like this: Servo[0,2] for a servo
        object connected to the first digital module and PWM port 2.

        :param moduleType: A string that defines the module name in the label
            for the value
        :param moduleNumber: The number of the particular module type
        :param channel: The channel number the device is plugged into
            (usually PWM)
        :param component: The reference to the object being added
        """
        LiveWindow.addActuator(
                "Ungrouped",
                "%s[%s,%s]" % (moduleType, moduleNumber, channel),
                component)

    @staticmethod
    def removeComponent(component):
        """Removes a component from LiveWindow.

        :param component: The reference to the object being removed.
        """
        if component in LiveWindow.components:
            if LiveWindow.components[component].isSensor:
                LiveWindow.sensors.remove(component)
            del(LiveWindow.components[component])



# validated: 2016-12-20 JW a705eb1c61a0 shared/java/edu/wpi/first/wpilibj/buttons/Trigger.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["Trigger"]

class Trigger:
    """This class provides an easy way to link commands to inputs.

    It is very easy to link a button to a command.  For instance, you could
    link the trigger button of a joystick to a "score" command.

    It is encouraged that teams write a subclass of Trigger if they want to
    have something unusual (for instance, if they want to react to the user
    holding a button while the robot is reading a certain sensor input).
    For this, they only have to write the :func:`get` method to get the full
    functionality of the Trigger class.
    """
    def get(self):
        """Returns whether or not the trigger is active

        This method will be called repeatedly a command is linked to the
        Trigger.

        :returns: whether or not the trigger condition is active.
        """
        raise NotImplementedError

    def grab(self):
        """Returns whether :meth:`get` returns True or the internal table for
        :class:`.SmartDashboard` use is pressed.
        """
        table = self.getTable()
        return self.get() or (table is not None and
                              table.getBoolean("pressed", False))

    def whenActive(self, command):
        """Starts the given command whenever the trigger just becomes active.

        :param command: the command to start
        """
        def execute():
            if self.grab():
                if not execute.pressedLast:
                    execute.pressedLast = True
                    command.start()
            else:
                execute.pressedLast = False

        execute.pressedLast = self.grab()
        from ..command import Scheduler
        Scheduler.getInstance().addButton(execute)

    def whileActive(self, command):
        """Constantly starts the given command while the button is held.

        :meth:`Command.start` will be called repeatedly while the trigger is
        active, and will be canceled when the trigger becomes inactive.

        :param command: the command to start
        """
        def execute():
            if self.grab():
                execute.pressedLast = True
                command.start()
            else:
                if execute.pressedLast:
                    execute.pressedLast = False
                    command.cancel()

        execute.pressedLast = self.grab()
        from ..command import Scheduler
        Scheduler.getInstance().addButton(execute)

    def whenInactive(self, command):
        """Starts the command when the trigger becomes inactive.

        :param command: the command to start
        """
        def execute():
            if self.grab():
                execute.pressedLast = True
            else:
                if execute.pressedLast:
                    execute.pressedLast = False
                    command.start()

        execute.pressedLast = self.grab()
        from ..command import Scheduler
        Scheduler.getInstance().addButton(execute)

    def toggleWhenActive(self, command):
        """Toggles a command when the trigger becomes active.

        :param command: the command to toggle
        """
        def execute():
            if self.grab():
                if not execute.pressedLast:
                    execute.pressedLast = True
                    if command.isRunning():
                        command.cancel()
                    else:
                        command.start()
            else:
                execute.pressedLast = False

        execute.pressedLast = self.grab()
        from ..command import Scheduler
        Scheduler.getInstance().addButton(execute)

    def cancelWhenActive(self, command):
        """Cancels a command when the trigger becomes active.

        :param command: the command to cancel
        """
        def execute():
            if self.grab():
                if not execute.pressedLast:
                    execute.pressedLast = True
                    command.cancel()
            else:
                execute.pressedLast = False

        execute.pressedLast = self.grab()
        from ..command import Scheduler
        Scheduler.getInstance().addButton(execute)

    def getSmartDashboardType(self):
        """These methods continue to return the "Button" :class:`.SmartDashboard` type
        until we decided to create a Trigger widget type for the dashboard.
        """
        return "Button"

    def initTable(self, table):
        self.table = table
        if table is not None:
            table.putBoolean("pressed", self.get())

    def getTable(self):
        return getattr(self, "table", None)



# validated: 2016-12-31 JW 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/Ultrasonic.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import threading
import weakref

from .counter import Counter
from .interfaces import PIDSource
from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase
from .timer import Timer

__all__ = ["Ultrasonic"]

class Ultrasonic(SensorBase):
    """Ultrasonic rangefinder control
    
    The Ultrasonic rangefinder measures
    absolute distance based on the round-trip time of a ping generated by
    the controller.  These sensors use two transducers, a speaker and a
    microphone both tuned to the ultrasonic range. A common ultrasonic
    sensor, the Daventech SRF04 requires a short pulse to be generated on
    a digital channel. This causes the chirp to be emmitted. A second line
    becomes high as the ping is transmitted and goes low when the echo is
    received. The time that the line is high determines the round trip
    distance (time of flight).
    
    .. not_implemented: initialize
    """

    class Unit:
        """The units to return when PIDGet is called"""
        kInches = 0
        kMillimeters = 1

    #: Time (sec) for the ping trigger pulse.
    kPingTime = 10 * 1e-6
    
    #: Priority that the ultrasonic round robin task runs.
    kPriority = 90
    
    #: Max time (ms) between readings.
    kMaxUltrasonicTime = 0.1
    kSpeedOfSoundInchesPerSec = 1130.0 * 12.0
    
    PIDSourceType = PIDSource.PIDSourceType

    _static_mutex = threading.RLock()
    
    #: ultrasonic sensor list
    sensors = weakref.WeakSet()
    
    #: Automatic round robin mode
    automaticEnabled = False
    instances = 0
    _thread = None

    @staticmethod
    def isAutomaticMode():
        with Ultrasonic._static_mutex:
            return Ultrasonic.automaticEnabled

    @staticmethod
    def ultrasonicChecker():
        """Background task that goes through the list of ultrasonic sensors
        and pings each one in turn. The counter is configured to read the
        timing of the returned echo pulse.

        .. warning:: DANGER WILL ROBINSON, DANGER WILL ROBINSON: This code runs
            as a task and assumes that none of the ultrasonic sensors will
            change while it's running. If one does, then this will certainly
            break. Make sure to disable automatic mode before changing
            anything with the sensors!!
        """
        while Ultrasonic.isAutomaticMode():
            count = 0
            for u in Ultrasonic.sensors:
                if not Ultrasonic.isAutomaticMode():
                    return
                if u is None:
                    continue
                count += 1
                if u.isEnabled():
                    # do the ping
                    u.pingChannel.pulse(Ultrasonic.kPingTime)
                Timer.delay(.1) # wait for ping to return
            if not count:
                return

    def __init__(self, pingChannel, echoChannel, units=Unit.kInches):
        """Create an instance of the Ultrasonic Sensor.
        This is designed to supchannel the Daventech SRF04 and Vex ultrasonic
        sensors.

        :param pingChannel: The digital output channel that sends the pulse
            to initiate the sensor sending the ping.
        :param echoChannel: The digital input channel that receives the echo.
            The length of time that the echo is high represents the round
            trip time of the ping, and the distance.
        :param units: The units returned in either kInches or kMillimeters
        """
        # Convert to DigitalInput and DigitalOutput if necessary
        self.pingAllocated = False
        self.echoAllocated = False
        
        if not hasattr(pingChannel, 'channel'):
            from .digitaloutput import DigitalOutput
            pingChannel = DigitalOutput(pingChannel)
            self.pingAllocated = True
            
        if not hasattr(echoChannel, 'channel'):
            from .digitalinput import DigitalInput
            echoChannel = DigitalInput(echoChannel)
            self.echoAllocated = True
            
        self.pingChannel = pingChannel
        self.echoChannel = echoChannel
        self.units = units
        self.pidSource = self.PIDSourceType.kDisplacement
        self.enabled = True # make it available for round robin scheduling
        
        # set up counter for this sensor
        self.counter = Counter(self.echoChannel)
        self.counter.setMaxPeriod(1.0)
        self.counter.setSemiPeriodMode(True)
        self.counter.reset()
        
        isAutomatic = Ultrasonic.isAutomaticMode()
        self.setAutomaticMode(False)
        
        Ultrasonic.sensors.add(self)
        if isAutomatic:
            self.setAutomaticMode(True)
        
        Resource._add_global_resource(self)

        Ultrasonic.instances += 1
        hal.report(hal.UsageReporting.kResourceType_Ultrasonic,
                      Ultrasonic.instances)
        LiveWindow.addSensor("Ultrasonic", self.echoChannel.getChannel(), self)

    def free(self):
        isAutomatic = Ultrasonic.isAutomaticMode()
        self.setAutomaticMode(False)
            
        try:
            Ultrasonic.sensors.remove(self)
        except (KeyError, ValueError):
            pass
        
        if isAutomatic and len(Ultrasonic.sensors):
            self.setAutomaticMode(True)
        
        if self.pingAllocated and self.pingChannel:
            self.pingChannel.free()
            self.pingChannel = None
            
        if self.echoAllocated and self.echoChannel:
            self.echoChannel.free()
            self.echoChannel = None
        
        if self.counter != None:
            self.counter.free()
            self.counter = None
        
        LiveWindow.removeComponent(self)
        super().free()

    def setAutomaticMode(self, enabling):
        """Turn Automatic mode on/off. When in Automatic mode, all sensors
        will fire in round robin, waiting a set time between each sensor.

        :param enabling:
            Set to true if round robin scheduling should start for all the
            ultrasonic sensors. This scheduling method assures that the
            sensors are non-interfering because no two sensors fire at the
            same time. If another scheduling algorithm is preferred, it
            can be implemented by pinging the sensors manually and waiting
            for the results to come back.
        :type enabling: bool
        """
        enabling = bool(enabling)
        if enabling == Ultrasonic.isAutomaticMode():
            return # ignore the case of no change
        
        with Ultrasonic._static_mutex:
            Ultrasonic.automaticEnabled = enabling

        if enabling:
            # Clear all the counters so no data is valid. No synchronization is
            # needed because the background task is stopped.
            for u in Ultrasonic.sensors:
                if u is not None:
                    u.counter.reset()
            
            # Start round robin task
            Ultrasonic._thread = threading.Thread(
                    target=Ultrasonic.ultrasonicChecker,
                    name="ultrasonicChecker")
            Ultrasonic.daemon = True
            Ultrasonic._thread.start()
        else:
            # Wait for background task to stop running
            Ultrasonic._thread.join()
            Ultrasonic._thread = None
            
            # Clear all the counters (data now invalid) since automatic mode is
            # disabled. No synchronization is needed because the background task is
            # stopped.
            for u in Ultrasonic.sensors:
                if u is not None:
                    u.counter.reset()

    def ping(self):
        """Single ping to ultrasonic sensor. Send out a single ping to the
        ultrasonic sensor. This only works if automatic (round robin) mode is
        disabled. A single ping is sent out, and the counter should count the
        semi-period when it comes in. The counter is reset to make the current
        value invalid.
        """
        # turn off automatic round robin if pinging single sensor
        self.setAutomaticMode(False)
        # reset the counter to zero (invalid data now)
        self.counter.reset()
        # do the ping to start getting a single range
        self.pingChannel.pulse(Ultrasonic.kPingTime)

    def isRangeValid(self):
        """Check if there is a valid range measurement. The ranges are
        accumulated in a counter that will increment on each edge of the
        echo (return) signal. If the count is not at least 2, then the range
        has not yet been measured, and is invalid.

        :returns: True if the range is valid
        :rtype: bool
        """
        return self.counter.get() > 1

    def getRangeInches(self):
        """Get the range in inches from the ultrasonic sensor.

        :returns: Range in inches of the target returned from the ultrasonic
            sensor. If there is no valid value yet, i.e. at least one
            measurement hasn't completed, then return 0.
        :rtype: float
        """
        if self.isRangeValid():
            return self.counter.getPeriod() * \
                    Ultrasonic.kSpeedOfSoundInchesPerSec / 2.0
        else:
            return 0

    def getRangeMM(self):
        """Get the range in millimeters from the ultrasonic sensor.

        :returns: Range in millimeters of the target returned by the
            ultrasonic sensor. If there is no valid value yet, i.e. at least
            one measurement hasn't complted, then return 0.
        :rtype: float
        """
        return self.getRangeInches() * 25.4
    
    def setPIDSourceType(self, pidSource):
        """Set which parameter you are using as a process
        control variable. 

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`.PIDSource.PIDSourceType`
        """
        if pidSource != self.PIDSourceType.kDisplacement:
            raise ValueError("Only displacement PID is allowed for ultrasonics.")
        self.pidSource = pidSource
        
    def getPIDSourceType(self):
        return self.pidSource

    def pidGet(self):
        """Get the range in the current DistanceUnit (PIDSource interface).

        :returns: The range in DistanceUnit
        :rtype: float
        """
        if self.units == self.Unit.kInches:
            return self.getRangeInches()
        elif self.units == self.Unit.kMillimeters:
            return self.getRangeMM()
        else:
            return 0.0

    def setDistanceUnits(self, units):
        """Set the current DistanceUnit that should be used for the
        PIDSource interface.

        :param units: The DistanceUnit that should be used.
        """
        
        if units not in [self.Unit.kInches, self.Unit.kMillimeters]:
            raise ValueError("Invalid units argument '%s'" % units)
        
        self.units = units

    def getDistanceUnits(self):
        """Get the current DistanceUnit that is used for the PIDSource
        interface.

        :returns: The type of DistanceUnit that is being used.
        """
        return self.units

    def isEnabled(self):
        """Is the ultrasonic enabled.

        :returns: True if the ultrasonic is enabled
        """
        return self.enabled

    def setEnabled(self, enable):
        """Set if the ultrasonic is enabled.

        :param enable: set to True to enable the ultrasonic
        :type  enable: bool
        """
        self.enabled = bool(enable)

    # Live Window code, only does anything if live window is activated.
    def getSmartDashboardType(self):
        return "Ultrasonic"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getRangeInches())

    def startLiveWindowMode(self):
        pass

    def stopLiveWindowMode(self):
        pass



# validated: 2016-11-15 AA 776cb91 shared/java/edu/wpi/first/wpilibj/PIDController.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from collections import deque
from itertools import islice
import threading
import warnings

import hal

from .interfaces import PIDSource
from .livewindowsendable import LiveWindowSendable
from .resource import Resource
from .timer import Timer
from ._impl.timertask import TimerTask
from ._impl.utils import match_arglist, HasAttribute

__all__ = ["PIDController"]

class PIDController(LiveWindowSendable):
    """Can be used to control devices via a PID Control Loop.

    Creates a separate thread which reads the given :class:`.PIDSource` and takes
    care of the integral calculations, as well as writing the given
    :class:`.PIDOutput`.

    This feedback controller runs in discrete time, so time deltas are not used
    in the integral and derivative calculations. Therefore, the sample rate affects
    the controller's behavior for a given set of PID constants.
    """
    kDefaultPeriod = .05
    instances = 0
    
    PIDSourceType = PIDSource.PIDSourceType

    # Tolerance is the type of tolerance used to specify if the PID controller
    # is on target.  The various implementations of this such as
    # PercentageTolerance and AbsoluteTolerance specify types of tolerance
    # specifications to use.
    def PercentageTolerance_onTarget(self, percentage):
        with self.mutex:
            return self.isAvgErrorValid() and \
                    (abs(self.getAvgError()) < percentage / 100.0
                    * (self.maximumInput - self.minimumInput))

    def AbsoluteTolerance_onTarget(self, value):
        return self.isAvgErrorValid() and abs(self.getAvgError()) < value

    def __init__(self, *args, **kwargs):
        """Allocate a PID object with the given constants for P, I, D, and F

        Arguments can be structured as follows:

        - Kp, Ki, Kd, Kf, PIDSource, PIDOutput, period
        - Kp, Ki, Kd, PIDSource, PIDOutput, period
        - Kp, Ki, Kd, PIDSource, PIDOutput
        - Kp, Ki, Kd, Kf, PIDSource, PIDOutput

        :param Kp: the proportional coefficient
        :type  Kp: float or int
        :param Ki: the integral coefficient
        :type  Ki: float or int
        :param Kd: the derivative coefficient
        :type  Kd: float or int
        :param Kf: the feed forward term
        :type  Kf: float or int
        :param source: Called to get values
        :type  source: A function, or an object that implements :class:`.PIDSource`
        :param output: Receives the output percentage
        :type  output: A function, or an object that implements :class:`.PIDOutput`
        :param period: the loop time for doing calculations. This particularly
            effects calculations of the integral and differential terms.
            The default is 50ms.
        :type  period: float or int
        """

        p_arg = ("Kp", [float, int])
        i_arg = ("Ki", [float, int])
        d_arg = ("Kd", [float, int])
        f_arg = ("kf", [float, int])
        source_arg = ("source", [HasAttribute("pidGet"), HasAttribute("__call__")])
        output_arg = ("output", [HasAttribute("pidWrite"), HasAttribute("__call__")])
        period_arg = ("period", [float, int])

        templates = [[p_arg, i_arg, d_arg, f_arg, source_arg, output_arg, period_arg],
                     [p_arg, i_arg, d_arg, source_arg, output_arg, period_arg],
                     [p_arg, i_arg, d_arg, source_arg, output_arg],
                     [p_arg, i_arg, d_arg, f_arg, source_arg, output_arg]]

        _, results = match_arglist('PIDController.__init__',
                                   args, kwargs, templates)

        self.P = results.pop("Kp")     # factor for "proportional" control
        self.I = results.pop("Ki")     # factor for "integral" control
        self.D = results.pop("Kd")     # factor for "derivative" control
        self.F = results.pop("Kf", 0.0)# factor for feedforward term
        self.pidOutput = results.pop("output")
        self.pidInput = results.pop("source")
        self.period = results.pop("period", PIDController.kDefaultPeriod)
        
        self.pidInput = PIDSource.from_obj_or_callable(self.pidInput)
        
        if hasattr(self.pidOutput, 'pidWrite'):
            self.pidOutput = self.pidOutput.pidWrite

        self.maximumOutput = 1.0    # |maximum output|
        self.minimumOutput = -1.0   # |minimum output|
        self.maximumInput = 0.0     # maximum input - limit setpoint to this
        self.minimumInput = 0.0     # minimum input - limit setpoint to this
        self.continuous = False     # do the endpoints wrap around? eg. Absolute encoder
        self.enabled = False        # is the pid controller enabled
        self.prevError = 0.0        # the prior error (used to compute velocity)
        self.totalError = 0.0       #the sum of the errors for use in the integral calc
        self.buf = deque(maxlen=1)
        self.setpoint = 0.0
        self.prevSetpoint = 0.0
        self.error = 0.0
        self.result = 0.0

        self.mutex = threading.RLock()

        self.pid_task = TimerTask('PIDTask%d' % PIDController.instances, self.period, self._calculate)
        self.pid_task.start()
        
        self.setpointTimer = Timer()
        self.setpointTimer.start()
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        PIDController.instances += 1
        hal.report(hal.UsageReporting.kResourceType_PIDController,
                   PIDController.instances)

    def free(self):
        """Free the PID object"""
        # TODO: is this useful in Python?  Should make TableListener weakref.
        self.pid_task.cancel()
        with self.mutex:
            self.pidInput = None
            self.pidOutput = None

    def _calculate(self):
        """Read the input, calculate the output accordingly, and write to the
        output.  This should only be called by the PIDTask and is created
        during initialization."""
        with self.mutex:
            if self.pidInput is None:
                return
            if self.pidOutput is None:
                return
            enabled = self.enabled # take snapshot of these values...
            pidInput = self.pidInput

        if enabled:
            with self.mutex:
                input = pidInput.pidGet()
                self.error = self.getContinuousError(self.setpoint - input)

                if self.pidInput.getPIDSourceType() == self.PIDSourceType.kRate:
                    if self.P != 0:
                        potentialPGain = (self.totalError + self.error) * self.P
                        if potentialPGain < self.maximumOutput:
                            if potentialPGain > self.minimumOutput:
                                self.totalError += self.error
                            else:
                                self.totalError = self.minimumOutput / self.P
                        
                        else:
                            self.totalError = self.maximumOutput / self.P
                        
                    self.result = self.P * self.totalError + \
                        self.D * self.error + \
                        self.calculateFeedForward()

                else:

                    if self.I != 0:
                        potentialIGain = (self.totalError + self.error) * self.I
                        if potentialIGain < self.maximumOutput:
                            if potentialIGain > self.minimumOutput:
                                self.totalError += self.error
                            else:
                                self.totalError = self.minimumOutput / self.I
                        else:
                            self.totalError = self.maximumOutput / self.I

                    self.result = self.P * self.error + \
                            self.I * self.totalError + \
                            self.D * (self.error - self.prevError) + \
                            self.calculateFeedForward()
                            
                self.prevError = self.error

                if self.result > self.maximumOutput:
                    self.result = self.maximumOutput
                elif self.result < self.minimumOutput:
                    self.result = self.minimumOutput
                pidOutput = self.pidOutput
                result = self.result
                
                self.buf.append(self.error)

            pidOutput(result)
            
    def calculateFeedForward(self):
        """Calculate the feed forward term
        
        Both of the provided feed forward calculations are velocity feed forwards.
        If a different feed forward calculation is desired, the user can override
        this function and provide his or her own. This function  does no
        synchronization because the PIDController class only calls it in
        synchronized code, so be careful if calling it oneself.
        
        If a velocity PID controller is being used, the F term should be set to 1
        over the maximum setpoint for the output. If a position PID controller is
        being used, the F term should be set to 1 over the maximum speed for the
        output measured in setpoint units per this controller's update period (see
        the default period in this class's constructor).
        """
        if self.pidInput.getPIDSourceType() == self.PIDSourceType.kRate:
            return self.F * self.getSetpoint()
        else:
            temp = self.F * self.getDeltaSetpoint()
            self.prevSetpoint = self.setpoint
            self.setpointTimer.reset()
            return temp

    def setPID(self, p, i, d, f=0.0):
        """Set the PID Controller gain parameters.
        Set the proportional, integral, and differential coefficients.

        :param p: Proportional coefficient
        :param i: Integral coefficient
        :param d: Differential coefficient
        :param f: Feed forward coefficient (optional, default is 0.0)
        """
        with self.mutex:
            self.P = p
            self.I = i
            self.D = d
            self.F = f

        table = self.getTable()
        if table is not None:
            table.putNumber("p", p)
            table.putNumber("i", i)
            table.putNumber("d", d)
            table.putNumber("f", f)

    def getP(self):
        """Get the Proportional coefficient.

        :returns: proportional coefficient
        """
        with self.mutex:
            return self.P

    def getI(self):
        """Get the Integral coefficient

        :returns: integral coefficient
        """
        with self.mutex:
            return self.I

    def getD(self):
        """Get the Differential coefficient.

        :returns: differential coefficient
        """
        with self.mutex:
            return self.D

    def getF(self):
        """Get the Feed forward coefficient.

        :returns: feed forward coefficient
        """
        with self.mutex:
            return self.F

    def get(self):
        """Return the current PID result.
        This is always centered on zero and constrained the the max and min
        outs.

        :returns: the latest calculated output
        """
        with self.mutex:
            return self.result

    def setContinuous(self, continuous=True):
        """Set the PID controller to consider the input to be continuous.
        Rather then using the max and min in as constraints, it considers them
        to be the same point and automatically calculates the shortest route
        to the setpoint.

        :param continuous: Set to True turns on continuous, False turns off
            continuous
        """
        with self.mutex:
            self.continuous = continuous

    def setInputRange(self, minimumInput, maximumInput):
        """Sets the maximum and minimum values expected from the input.
        
        :param minimumInput: the minimum percentage expected from the input
        :param maximumInput: the maximum percentage expected from the output
        """
        with self.mutex:
            if minimumInput > maximumInput:
                raise ValueError("Lower bound is greater than upper bound")
            self.minimumInput = minimumInput
            self.maximumInput = maximumInput
            self.setSetpoint(self.setpoint)

    def setOutputRange(self, minimumOutput, maximumOutput):
        """Sets the minimum and maximum values to write.

        :param minimumOutput: the minimum percentage to write to the output
        :param maximumOutput: the maximum percentage to write to the output
        """
        with self.mutex:
            if minimumOutput > maximumOutput:
                raise ValueError("Lower bound is greater than upper bound")
            self.minimumOutput = minimumOutput
            self.maximumOutput = maximumOutput

    def setSetpoint(self, setpoint):
        """Set the setpoint for the PIDController Clears the queue for GetAvgError().

        :param setpoint: the desired setpoint
        """
        with self.mutex:
            if self.maximumInput > self.minimumInput:
                if setpoint > self.maximumInput:
                    newsetpoint = self.maximumInput
                elif setpoint < self.minimumInput:
                    newsetpoint = self.minimumInput
                else:
                    newsetpoint = setpoint
            else:
                newsetpoint = setpoint
            self.setpoint = newsetpoint
            
            self.buf.clear()
            self.totalError = 0

        table = self.getTable()
        if table is not None:
            table.putNumber("setpoint", newsetpoint)

    def getSetpoint(self):
        """Returns the current setpoint of the PIDController.

        :returns: the current setpoint
        """
        with self.mutex:
            return self.setpoint
        
    def getDeltaSetpoint(self):
        """Returns the change in setpoint over time of the PIDController
        
        :returns: the change in setpoint over time
        """
        with self.mutex:
            t = self.setpointTimer.get()
            # During testing/simulation it is possible to get a divide by zero because
            # the threads' calls aren't strictly aligned with the master clock.
            if t:
                return (self.setpoint - self.prevSetpoint) / t
            else:
                return 0.0

    def getError(self):
        """Returns the current difference of the input from the setpoint.

        :return: the current error
        """
        with self.mutex:
            #return self.error
            return self.getContinuousError(self.getSetpoint() - self.pidInput())
        
    def setPIDSourceType(self, pidSourceType):
        """Sets what type of input the PID controller will use
        
        :param pidSourceType: the type of input
        """
        self.pidInput.setPIDSourceType(pidSourceType)
        
    def getPIDSourceType(self, pidSourceType):
        """Returns the type of input the PID controller is using
        
        :returns: the PID controller input type
        """
        return self.pidInput.getPIDSourceType()
    
    def getAvgError(self):
        """Returns the current difference of the error over the past few iterations.
        You can specify the number of iterations to average with
        :meth:`setToleranceBuffer` (defaults to 1). getAvgError() is used for the
        onTarget() function.
        
        :returns: the current average of the error
        """
        with self.mutex:
            l = len(self.buf)
            if l == 0:
                return 0
            else:
                return sum(self.buf) / l
            
    def isAvgErrorValid(self):
        """Returns whether or not any values have been collected. If no values
        have been collected, getAvgError is 0, which is invalid.
        
        :returns: True if :meth:`getAvgError` is currently valid.
        """
        with self.mutex:
            return len(self.buf) != 0

    def setTolerance(self, percent):
        """Set the percentage error which is considered tolerable for use with
        :func:`onTarget`. (Input of 15.0 = 15 percent)

        :param percent: error which is tolerable

        .. deprecated:: 2015.1

            Use :func:`setPercentTolerance` or :func:`setAbsoluteTolerance`
            instead.
        """
        warnings.warn("use setPercentTolerance or setAbsoluteTolerance instead",
                      DeprecationWarning)
        with self.mutex:
            self.onTarget = lambda: \
                    self.PercentageTolerance_onTarget(percent)

    def setAbsoluteTolerance(self, absvalue):
        """Set the absolute error which is considered tolerable for use with
        :func:`onTarget`.

        :param absvalue: absolute error which is tolerable in the units of the
            input object
        """
        with self.mutex:
            self.onTarget = lambda: \
                    self.AbsoluteTolerance_onTarget(absvalue)

    def setPercentTolerance(self, percentage):
        """Set the percentage error which is considered tolerable for use with
        :func:`onTarget`. (Input of 15.0 = 15 percent)

        :param percentage: percent error which is tolerable
        """
        with self.mutex:
            self.onTarget = lambda: \
                    self.PercentageTolerance_onTarget(percentage)
                    
    def setToleranceBuffer(self, bufLength):
        """Set the number of previous error samples to average for tolerancing. When
        determining whether a mechanism is on target, the user may want to use a
        rolling average of previous measurements instead of a precise position or
        velocity. This is useful for noisy sensors which return a few erroneous
        measurements when the mechanism is on target. However, the mechanism will
        not register as on target for at least the specified bufLength cycles.
        
        :param bufLength: Number of previous cycles to average.
        :type bufLength: int
        """
        with self.mutex:
            self.buf = deque(list(islice(self.buf, bufLength)), maxlen=bufLength)
        
    def onTarget(self):
        """Return True if the error is within the percentage of the total input
        range, determined by setTolerance. This assumes that the maximum and
        minimum input were set using :func:`setInput`.

        :returns: True if the error is less than the tolerance
        """
        raise ValueError("No tolerance value set when using PIDController.onTarget()")

    def enable(self):
        """Begin running the PIDController."""
        with self.mutex:
            self.enabled = True

        table = self.getTable()
        if table is not None:
            table.putBoolean("enabled", True)

    def disable(self):
        """Stop running the PIDController, this sets the output to zero before
        stopping."""
        with self.mutex:
            self.pidOutput(0)
            self.enabled = False

        table = self.getTable()
        if table is not None:
            table.putBoolean("enabled", False)

    def isEnable(self):
        """Return True if PIDController is enabled."""
        with self.mutex:
            return self.enabled

    def reset(self):
        """Reset the previous error, the integral term, and disable the
        controller."""
        with self.mutex:
            self.disable()
            self.prevError = 0
            self.totalError = 0
            self.result = 0

    def getSmartDashboardType(self):
        return "PIDController"

    def valueChanged(self, table, key, value, isNew):
        if key == "p" or key == "i" or key == "d" or key == "f":
            Kp = table.getNumber("p", 0.0)
            Ki = table.getNumber("i", 0.0)
            Kd = table.getNumber("d", 0.0)
            Kf = table.getNumber("f", 0.0)
            if (self.getP() != Kp or self.getI() != Ki or self.getD() != Kd or
                self.getF() != Kf):
                self.setPID(Kp, Ki, Kd, Kf)
        elif key == "setpoint":
            if self.getSetpoint() != float(value):
                self.setSetpoint(float(value))
        elif key == "enabled":
            if self.isEnable() != bool(value):
                if bool(value):
                    self.enable()
                else:
                    self.disable()

    def initTable(self, table):
        oldtable = self.getTable()
        if oldtable is not None:
            oldtable.removeTableListener(self.valueChanged)
        self.table = table
        if table is not None:
            table.putNumber("p", self.getP())
            table.putNumber("i", self.getI())
            table.putNumber("d", self.getD())
            table.putNumber("f", self.getF())
            table.putNumber("setpoint", self.getSetpoint())
            table.putBoolean("enabled", self.isEnable())
            table.addTableListener(self.valueChanged, False)

    def getContinuousError(self, error):
        """
        Wraps error around for continuous inputs. The original error is
        returned if continuous mode is disabled. This is an unsynchronized
        function.

        :param error: The current error of the PID controller.
        :return: Error for continuous inputs.
        """
        if self.continuous:
            if abs(error) > (self.maximumInput - self.minimumInput) / 2:
                if error > 0:
                    return error - (self.maximumInput - self.minimumInput)
                else:
                    return error + (self.maximumInput - self.minimumInput)
        return error

    def startLiveWindowMode(self):
        self.disable()

    def stopLiveWindowMode(self):
        pass



# novalidate
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import logging

from networktables import NetworkTables

import hal


logger = logging.getLogger(__name__)

__all__ = ["Preferences"]


class Preferences:
    """Provides a relatively simple way to save important
    values to the roboRIO to access the next time the roboRIO is booted.

    This class loads and saves from a file inside the roboRIO. The user can not
    access the file directly, but may modify values at specific fields which
    will then be saved to the file when :func:`save` is called.

    This class is thread safe.

    This will also interact with :class:`networktables.NetworkTable` by creating a table
    called "Preferences" with all the key-value pairs. To save using
    `NetworkTable`, simply set the boolean at position ~S A V E~ to true.
    Also, if the value of any variable is " in the `NetworkTable`, then
    that represents non-existence in the `Preferences` table.

    .. not_implemented: putDouble, putLong, getDouble, getLong
    """

    # The Preferences table name
    TABLE_NAME = "Preferences"

    @staticmethod
    def getInstance():
        """Returns the preferences instance.

        :returns: the preferences instance
        """
        if not hasattr(Preferences, "instance"):
            Preferences.instance = Preferences()
        return Preferences.instance

    def __init__(self):
        """Creates a preference class that will automatically read the file in
        a different thread. Any call to its methods will be blocked until the
        thread is finished reading.
        """
        self.table = NetworkTables.getTable(self.TABLE_NAME)
        self.table.addTableListenerEx(self.valueChangedEx, NetworkTables.NotifyFlags.NEW | NetworkTables.NotifyFlags.IMMEDIATE)

        hal.report(hal.UsageReporting.kResourceType_Preferences, 0)

    def getKeys(self):
        """:returns: a list of the keys
        """
        with self.lock:
            return [x for x in self.table.getKeys()]

    def keys(self):
        """Python style get list of keys.
        """
        with self.lock:
            return [x for x in self.table.getKeys()]

    def putString(self, key, value):
        """Puts the given string into the preferences table.

        The value may not have quotation marks, nor may the key have any
        whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        if '"' in value:
            raise ValueError("Can not put string: '%s' because it contains quotation marks" % value)
        self.table.putString(key, value)
        self.table.setPersistent(key)

    def putInt(self, key, value):
        """Puts the given int into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.table.putNumber(key, value)
        self.table.setPersistent(key)

    def putFloat(self, key, value):
        """Puts the given float into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.table.putNumber(key, value)
        self.table.setPersistent(key)

    def putBoolean(self, key, value):
        """Puts the given float into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.table.putBoolean(key, value)
        self.table.setPersistent(key)

    def __setitem__(self, key, value):
        """Python style setting of key/value."""
        self.table.putString(key, str(value))
  
    def containsKey(self, key):
        """Returns whether or not there is a key with the given name.

        :param key: the key
        :returns: True if there is a value at the given key
        """
        return self.table.containsKey(key)

    def __contains__(self, key):
        """Python style contains key."""
        return self.table.containsKey(key)

    def remove(self, key):
        """Remove a preference

        :param key: the key
        """
        self.table.delete(key)

    def __delitem__(self, key):
        """Python style preference removal
        """
        self.table.delete(key)

    def getString(self, key, backup=None):
        """Returns the string at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        """
        return self.table.getString(key, backup)

    def getInt(self, key, backup=None):
        """Returns the int at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        :raises: TableKeyNotDefinedException if key cannot be found
        """
        return self.table.getNumber(key, backup)

    def getFloat(self, key, backup=None):
        """Returns the float at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        :raises: TableKeyNotDefinedException if key cannot be found
        """
        return self.table.getNumber(key, backup)

    def getBoolean(self, key, backup=None):
        """Returns the boolean at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        """
        return self.table.getBoolean(key, backup)

    def valueChangedEx(self, source, key, value, isNew):
        self.table.setPersistent(key)



# validated: 2016-12-25 JW 963391cf3916 athena/java/edu/wpi/first/wpilibj/SolenoidBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import threading

from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["SolenoidBase"]

class SolenoidBase(SensorBase):
    """SolenoidBase class is the common base class for the Solenoid and
    DoubleSolenoid classes."""

    # global to all instances, keyed by module number
    all_allocated = {}
    all_ports = {}
    all_mutex = {}
    
    @staticmethod
    def _reset():
        SolenoidBase.all_allocated = {}
        SolenoidBase.all_ports = {}
        SolenoidBase.all_mutex = {}

    def __init__(self, moduleNumber):
        """Constructor.

        :param moduleNumber: The PCM CAN ID
        """
        self.moduleNumber = moduleNumber

        if moduleNumber not in self.all_ports:
            self.all_ports[moduleNumber] = []

            for i in range(SensorBase.kSolenoidChannels):
                port = hal.getPortWithModule(moduleNumber, i)
                self.all_ports[moduleNumber].append(hal.initializeSolenoidPort(port))

        if moduleNumber not in self.all_mutex:
            self.all_mutex[moduleNumber] = threading.Lock()

        if moduleNumber not in self.all_allocated:
            self.all_allocated[moduleNumber] = Resource(SensorBase.kSolenoidChannels)

        self.allocated = self.all_allocated[moduleNumber]
        self.ports = self.all_ports[moduleNumber]
        self.mutex = self.all_mutex[moduleNumber]

    def getAll(self):
        """Read all 8 solenoids from the module used by this solenoid as a
        single byte.

        :returns: The current value of all 8 solenoids on this module.
        """
        with self.mutex:
            return hal.getAllSolenoids(self.moduleNumber)

    def getPCMSolenoidBlackList(self):
        """
        Reads complete solenoid blacklist for all 8 solenoids as a single byte.
            If a solenoid is shorted, it is added to the blacklist and
            disabled until power cycle, or until faults are cleared. See
            :meth:`clearAllPCMStickyFaults`

        :returns: The solenoid blacklist of all 8 solenoids on the module.
        """
        return hal.getPCMSolenoidBlackList(self.moduleNumber)

    def getPCMSolenoidVoltageStickyFault(self):
        """
        :returns: True if PCM Sticky fault is set : The common
            highside solenoid voltage rail is too low, most likely
            a solenoid channel has been shorted.
        """
        return hal.getPCMSolenoidVoltageStickyFault(self.moduleNumber)

    def getPCMSolenoidVoltageFault(self):
        """
        :returns: True if PCM is in fault state : The common
            highside solenoid voltage rail is too low, most likely
            a solenoid channel has been shorted.
        """
        return hal.getPCMSolenoidVoltageFault(self.moduleNumber)

    def clearAllPCMStickyFaults(self):
        """
        Clear ALL sticky faults inside the PCM that Solenoid is wired to.

        If a sticky fault is set, then it will be persistently cleared. Compressor drive
            maybe momentarily disable while flages are being cleared. Care should be
            taken to not call this too frequently, otherwise normal compressor functionality
            may be prevented.

        If no sticky faults are set then this call will have no effect.
        """
        hal.clearAllPCMStickyFaults(self.moduleNumber)




# validated: 2016-12-01 AA 140c365e4b99 shared/java/edu/wpi/first/wpilibj/JoystickBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .generichid import GenericHID

__all__ = ["JoystickBase"]

class JoystickBase(GenericHID):
    """
    JoystickBase Interface.
    """

    def __init__(self, port):
        super().__init__(port)

    def getZ(self, hand=None):
        """Get the z position of the HID.

        :param hand: which hand, left or right
        :returns: the z position
        """

        raise NotImplementedError

    def getTwist(self):
        """Get the twist value.

        :returns: the twist value
        """

        raise NotImplementedError

    def getThrottle(self):
        """Get the throttle.

        :returns: the throttle value
        """

        raise NotImplementedError

    def getTrigger(self, hand=None):
        """Is the trigger pressed.

        :param hand: which hand
        :returns: true if the trigger for the given hand is pressed
        """

        raise NotImplementedError

    def getTop(self, hand=None):
        """Is the top button pressed.

        :param hand: which hand
        :returns: true if hte top button for the given hand is pressed
        """

        raise NotImplementedError

    def getPOV(self, pov=0):
        raise NotImplementedError

    def getPOVCount(self):
        raise NotImplementedError

    def getType(self):
        raise NotImplementedError

    def getName(self):
        raise NotImplementedError

    def setOutput(self, outputNumber, value):
        raise NotImplementedError

    def setOutputs(self, value):
        raise NotImplementedError

    def setRumble(self, type, value):
        raise NotImplementedError



# validated: 2016-11-26 DS 952ebb11adfe shared/java/edu/wpi/first/wpilibj/MotorSafety.java athena/java/edu/wpi/first/wpilibj/MotorSafetyHelper.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import weakref
import threading

from .robotstate import RobotState
from .timer import Timer

import logging
logger = logging.getLogger(__name__)

__all__ = ["MotorSafety"]

class MotorSafety:
    """Provides mechanisms to safely shutdown motors if they aren't updated
    often enough.
    
    The MotorSafety object is constructed for every object that wants to
    implement the Motor Safety protocol. The helper object has the code to
    actually do the timing and call the motors stop() method when the timeout
    expires. The motor object is expected to call the feed() method whenever
    the motors value is updated.
    
    .. robotpy-specific:: In the Java implementation this is MotorSafetyHelper
    """
    DEFAULT_SAFETY_EXPIRATION = 0.1
    helpers = weakref.WeakSet()
    helpers_lock = threading.Lock()
    
    @staticmethod
    def _reset():
        with MotorSafety.helpers_lock:
            MotorSafety.helpers.clear()

    def __init__(self):
        """The constructor for a MotorSafety object.
        The helper object is constructed for every object that wants to
        implement the Motor Safety protocol. The helper object has the code
        to actually do the timing and call the motors stop() method when the
        timeout expires. The motor object is expected to call the feed()
        method whenever the motors value is updated.
        """
        self.safetyEnabled = False
        self.safetyExpiration = MotorSafety.DEFAULT_SAFETY_EXPIRATION
        self.safetyStopTime = Timer.getFPGATimestamp()
        with MotorSafety.helpers_lock:
            MotorSafety.helpers.add(self)

    def feed(self):
        """Feed the motor safety object.
        Resets the timer on this object that is used to do the timeouts.
        """
        self.safetyStopTime = Timer.getFPGATimestamp() + self.safetyExpiration

    def setExpiration(self, expirationTime):
        """Set the expiration time for the corresponding motor safety object.

        :param expirationTime: The timeout value in seconds.
        :type expirationTime: float
        """
        self.safetyExpiration = expirationTime

    def getExpiration(self):
        """Retrieve the timeout value for the corresponding motor safety
        object.

        :returns: the timeout value in seconds.
        :rtype: float
        """
        return self.safetyExpiration

    def isAlive(self):
        """Determine of the motor is still operating or has timed out.

        :returns: True if the motor is still operating normally and hasn't
            timed out.
        :rtype: float
        """
        return not self.safetyEnabled or self.safetyStopTime > Timer.getFPGATimestamp()

    def check(self):
        """Check if this motor has exceeded its timeout.
        This method is called periodically to determine if this motor has
        exceeded its timeout value. If it has, the stop method is called,
        and the motor is shut down until its value is updated again.
        """
        if not self.safetyEnabled or RobotState.isDisabled() or RobotState.isTest():
            return
        if self.safetyStopTime < Timer.getFPGATimestamp():
            logger.warn("%s... Output not updated often enough." %
                        self.getDescription())

            self.stopMotor()

    def setSafetyEnabled(self, enabled):
        """Enable/disable motor safety for this device.
        Turn on and off the motor safety option for this PWM object.

        :param enabled: True if motor safety is enforced for this object
        :type  enabled: bool
        """
        self.safetyEnabled = bool(enabled)

    def isSafetyEnabled(self):
        """Return the state of the motor safety enabled flag.
        Return if the motor safety is currently enabled for this device.

        :returns: True if motor safety is enforced for this device
        :rtype: bool
        """
        return self.safetyEnabled

    @staticmethod
    def checkMotors():
        """Check the motors to see if any have timed out.
        This static method is called periodically to poll all the motors and
        stop any that have timed out.
        """
        with MotorSafety.helpers_lock:
            for msh in MotorSafety.helpers:
                msh.check()



# validated: 2015-12-22 DS 9b4dd26 shared/java/edu/wpi/first/wpilibj/interfaces/Gyro.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.
#----------------------------------------------------------------------------

__all__ = ['Gyro']


class Gyro:
    """Interface for yaw rate gyros"""
    
    def calibrate(self):
        """
        Calibrate the gyro by running for a number of samples and computing the
        center value. Then use the center value as the Accumulator center value for
        subsequent measurements.
        
        .. note:: It's important to make sure that the robot is not
            moving while the centering calculations are in progress, this is typically
            done when the robot is first turned on while it's sitting at rest before
            the competition starts.
        """
        raise NotImplementedError

    def reset(self):
        """
        Reset the gyro. Resets the gyro to a heading of zero. This can be used if
        there is significant drift in the gyro and it needs to be recalibrated
        after it has been running.
        """
        raise NotImplementedError
    
    def getAngle(self):
        """
        Return the actual angle in degrees that the robot is currently facing.
        
        The angle is based on the current accumulator value corrected by the
        oversampling rate, the gyro type and the A/D calibration values. The angle
        is continuous, that is it will continue from 360 to 361 degrees. This
        allows algorithms that wouldn't want to see a discontinuity in the gyro
        output as it sweeps past from 360 to 0 on the second time around.
        
        :returns: the current heading of the robot in degrees. This heading is based
                  on integration of the returned rate from the gyro.
        """
        raise NotImplementedError

    def getRate(self):
        """
        Return the rate of rotation of the gyro
    
        The rate is based on the most recent reading of the gyro analog value
    
        :returns: the current rate in degrees per second
        """
        raise NotImplementedError
    
    def free(self):
        """Free the resources used by the gyro"""
        raise NotImplementedError



# validated: 2015-12-24 DS 6d854af shared/java/edu/wpi/first/wpilibj/RobotState.java

__all__ = ["RobotState"]

class RobotState:
    """Provides an interface to determine the current operating state of the
    robot code.
    """
    
    impl = None

    @staticmethod
    def isDisabled():
        return RobotState.impl.isDisabled()

    @staticmethod
    def isEnabled():
        return RobotState.impl.isEnabled()

    @staticmethod
    def isOperatorControl():
        return RobotState.impl.isOperatorControl()

    @staticmethod
    def isAutonomous():
        return RobotState.impl.isAutonomous()

    @staticmethod
    def isTest():
        return RobotState.impl.isTest()



# validated: 2016-11-20 KC b78f580d476f shared/java/edu/wpi/first/wpilibj/command/StartCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .instantcommand import InstantCommand

__all__ = ["StartCommand"]

class StartCommand(InstantCommand):
    """A StartCommand will call the start() method of another command when it
    is initialized and will finish immediately.
    """

    def __init__(self, commandToStart):
        """Instantiates a StartCommand which will start the
        given command whenever its initialize() is called.

        :param commandToStart: the :class:`.Command` to start
        """
        super().__init__("Start(%s)" % commandToStart)
        self.commandToFork = commandToStart

    def initialize(self):
        self.commandToFork.start()



# validated: 2016-11-20 KC b25a7cb3704d shared/java/edu/wpi/first/wpilibj/command/InstantCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------
from .command import Command

__all__ = ["InstantCommand"]

class InstantCommand(Command):
    '''
    A command that has no duration. Subclasses should implement the initialize()
    method to carry out desired actions.
    '''

    def __init__(self, name=None):
        super().__init__(name)


    def isFinished(self):
        return True




# validated: 2016-12-31 JW 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/ADXRS450_Gyro.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2015. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .driverstation import DriverStation
from .gyrobase import GyroBase
from .livewindow import LiveWindow
from .spi import SPI
from .timer import Timer
    

class ADXRS450_Gyro(GyroBase):
    """
    Use a rate gyro to return the robots heading relative to a starting position.
    The Gyro class tracks the robots heading based on the starting position. As
    the robot rotates the new heading is computed by integrating the rate of
    rotation returned by the sensor. When the class is instantiated, it does a
    short calibration routine where it samples the gyro while at rest to
    determine the default offset. This is subtracted from each sample to
    determine the heading.
    
    This class is for the digital ADXRS450 gyro sensor that connects via SPI.
    """
    
    kSamplePeriod = 0.001
    kCalibrationSampleTime = 5.0
    kDegreePerSecondPerLSB = 0.0125

    kRateRegister = 0x00
    kTemRegister = 0x02
    kLoCSTRegister = 0x04
    kHiCSTRegister = 0x06
    kQuadRegister = 0x08
    kFaultRegister = 0x0A
    kPIDRegister = 0x0C
    kSNHighRegister = 0x0E
    kSNLowRegister = 0x10

    def __init__(self, port=None):
        """
            Constructor.
        
            :param port: The SPI port that the gyro is connected to
            :type port: :class:`.SPI.Port`
        """
        
        if port is None:
            port = SPI.Port.kOnboardCS0
        
        simPort = None
        if hal.HALIsSimulation():
            from hal_impl.spi_helpers import ADXRS450_Gyro_Sim
            simPort = ADXRS450_Gyro_Sim(self)
        
        self.spi = SPI(port, simPort=simPort)
        self.spi.setClockRate(3000000)
        self.spi.setMSBFirst()
        self.spi.setSampleDataOnRising()
        self.spi.setClockActiveHigh()
        self.spi.setChipSelectActiveLow()

        # Validate the part ID
        if (self._readRegister(self.kPIDRegister) & 0xff00) != 0x5200:
            self.spi.free()
            self.spi = None
            DriverStation.reportError("could not find ADXRS450 gyro on SPI port %s" % port, False)
            return
        
        self.spi.initAccumulator(self.kSamplePeriod, 0x20000000, 4, 0x0c00000e, 0x04000000,
            10, 16, True, True)
        
        self.calibrate()
        
        hal.report(hal.UsageReporting.kResourceType_ADXRS450, port)
        LiveWindow.addSensor("ADXRS450_Gyro", port, self)

    def calibrate(self):
        if self.spi is None:
            return
        
        if not hal.HALIsSimulation():
            Timer.delay(0.1)

        self.spi.setAccumulatorCenter(0)
        self.spi.resetAccumulator()
        
        if not hal.HALIsSimulation():
            Timer.delay(self.kCalibrationSampleTime)
        
        self.spi.setAccumulatorCenter(int(self.spi.getAccumulatorAverage()))
        self.spi.resetAccumulator()

    def _calcParity(self, v):
        parity = False
        while v != 0:
            parity = not parity
            v = v & (v - 1)
        return parity
    
    def _readRegister(self, reg):
        cmdhi = 0x8000 | (reg << 1)
        parity = self._calcParity(cmdhi)
    
        data = [cmdhi >> 8,
                cmdhi & 0xff,
                0,
                0 if parity else 1] 
    
        self.spi.write(data)
        data = self.spi.read(False, 4)
    
        if (data[0] & 0xe0) == 0:
            return 0  # error, return 0
        
        val = int.from_bytes(data[:4], byteorder='big')
        return (val >> 5) & 0xffff
    
    def reset(self):
        self.spi.resetAccumulator()
        
    def free(self):
        """Delete (free) the spi port used for the gyro and stop accumulating."""
        if self.spi is not None:
            self.spi.free()
            self.spi = None
    
    def getAngle(self):
        if self.spi is None:
            return 0.0
        return self.spi.getAccumulatorValue() * self.kDegreePerSecondPerLSB * self.kSamplePeriod

    def getRate(self):
        if self.spi is None:
            return 0.0
        else:
            return self.spi.getAccumulatorLastValue() * self.kDegreePerSecondPerLSB




# validated: 2016-12-31 JW 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/DigitalOutput.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import warnings
import weakref

import hal

from .digitalsource import DigitalSource
from .sensorbase import SensorBase

__all__ = ["DigitalOutput"]

def _freePWMGenerator(pwmGenerator):
    # Disable the output by routing to a dead bit.
    hal.setDigitalPWMOutputChannel(pwmGenerator, SensorBase.kDigitalChannels)
    hal.freeDigitalPWM(pwmGenerator)

class DigitalOutput(DigitalSource):
    """Writes to a digital output
    
    Other devices that are implemented elsewhere will automatically allocate
    digital inputs and outputs as required.
    """

    def __init__(self, channel):
        """Create an instance of a digital output.

        :param channel: the DIO channel for the digital output. 0-9 are on-board, 10-25 are on the MXP
        """
        super().__init__(channel, False)
        self._pwmGenerator = None
        self._pwmGenerator_finalizer = None

        hal.report(hal.UsageReporting.kResourceType_DigitalOutput,
                      channel)

    @property
    def pwmGenerator(self):
        if self._pwmGenerator_finalizer is None:
            return None
        if not self._pwmGenerator_finalizer.alive:
            return None
        return self._pwmGenerator

    def free(self):
        """Free the resources associated with a digital output."""
        # finalize the pwm only if we have allocated it
        if self.pwmGenerator is not None:
            self._pwmGenerator_finalizer()
        self._pwmGenerator = None
        super().free()
        self._handle = None

    def set(self, value):
        """Set the value of a digital output.

        :param value: True is on, off is False
        :type  value: bool
        """
        hal.setDIO(self.port, 1 if value else 0)

    def get(self):
        """Gets the value being output from the Digital Output.

        :returns: the state of the digital output
        :rtype: bool
        """
        return hal.getDIO(self.handle)

    def getChannel(self):
        """:returns: The GPIO channel number that this object represents.
        """
        return self.channel

    def pulse(self, pulseLength, *args):
        """Generate a single pulse. There can only be a single pulse going at any time.

        :param channel: Unused. Deprecated 2017.1.1.
        :param pulseLength: The length of the pulse.
        :type  pulseLength: float
        """
        if len(args) > 0:
            warnings.warn("Use of channel argument is deprecated.", DeprecationWarning)
            pulseLength = args[0]
        hal.pulse(self.handle, pulseLength)

    def isPulsing(self):
        """Determine if the pulse is still going. Determine if a previously
        started pulse is still going.

        :returns: True if pulsing
        :rtype: bool
        """
        return hal.isPulsing(self.handle)

    def setPWMRate(self, rate):
        """Change the PWM frequency of the PWM output on a Digital Output line.

        The valid range is from 0.6 Hz to 19 kHz. The frequency resolution is
        logarithmic.

        There is only one PWM frequency for all channnels.

        :param rate: The frequency to output all digital output PWM signals.
        :type  rate: float
        """
        hal.setDigitalPWMRate(rate)

    def enablePWM(self, initialDutyCycle):
        """Enable a PWM Output on this line.

        Allocate one of the 6 DO PWM generator resources.

        Supply the initial duty-cycle to output so as to avoid a glitch when
        first starting.

        The resolution of the duty cycle is 8-bit for low frequencies (1kHz or
        less) but is reduced the higher the frequency of the PWM signal is.

        :param initialDutyCycle: The duty-cycle to start generating. [0..1]
        :type  initialDutyCycle: float
        """
        if self.pwmGenerator is not None:
            return
        self._pwmGenerator = hal.allocateDigitalPWM()
        hal.setDigitalPWMDutyCycle(self._pwmGenerator, initialDutyCycle)
        hal.setDigitalPWMOutputChannel(self._pwmGenerator, self.channel)
        self._pwmGenerator_finalizer = \
                weakref.finalize(self, _freePWMGenerator, self._pwmGenerator)

    def disablePWM(self):
        """Change this line from a PWM output back to a static Digital Output
        line.

        Free up one of the 6 DO PWM generator resources that were in use.
        """
        if self.pwmGenerator is None:
            return
        self._pwmGenerator_finalizer()

    def updateDutyCycle(self, dutyCycle):
        """Change the duty-cycle that is being generated on the line.

        The resolution of the duty cycle is 8-bit for low frequencies (1kHz or
        less) but is reduced the higher the frequency of the PWM signal is.

        :param dutyCycle: The duty-cycle to change to. [0..1]
        :type  dutyCycle: float
        """
        if self.pwmGenerator is None:
            return
        hal.setDigitalPWMDutyCycle(self.pwmGenerator, dutyCycle)

    def getAnalogTriggerTypeForRouting(self):
        """Get the analog trigger type.

        :returns: false
        :rtype: int
        """
        return 0

    def isAnalogTrigger(self):
        """Is this an analog trigger.

        :returns: true if this is an analog trigger
        :rtype: bool
        """
        return False

    def getPortHandleForRouting(self):
        """Get the HAL Port Handle.

        :returns: The HAL Handle to the specified source
        """
        return self.handle

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    # Live Window code, only does anything if live window is activated.
    def getSmartDashboardType(self):
        return "Digital Output"

    def updateTable(self):
        # TODO: Put current value.
        pass

    def valueChanged(self, itable, key, value, bln):
        self.set(True if value else False)



# validated: 2016-11-20 KC b78f580d476f shared/java/edu/wpi/first/wpilibj/command/WaitCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .timedcommand import TimedCommand

__all__ = ["WaitCommand"]

class WaitCommand(TimedCommand):
    """A WaitCommand will wait for a certain amount of time before finishing.
    It is useful if you want a :class:`.CommandGroup` to pause for a moment.
    
    .. seealso:: :class:`.CommandGroup`
    """

    def __init__(self, timeout, name=None):
        """Instantiates a WaitCommand with the given timeout.
        
        :param timeout: the time the command takes to run
        :param name: the name of the command (optional)
        """
        if name is None:
            super().__init__("Wait(%s)" % timeout, timeout)
        else:
            super().__init__(name, timeout)



# validated: 2016-11-26 DS 5ad28d58ec63 athena/java/edu/wpi/first/wpilibj/SafePWM.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .motorsafety import MotorSafety
from .pwm import PWM

__all__ = ["SafePWM"]

class SafePWM(PWM, MotorSafety):
    """
        A raw PWM interface that implements the :class:`.MotorSafety` interface
    """
    
    def __init__(self, channel):
        """Constructor for a SafePWM object taking a channel number.

        :param channel: The channel number to be used for the underlying PWM
            object. 0-9 are on-board, 10-19 are on the MXP port.
        :type  channel: int
        """
        MotorSafety.__init__(self)
        PWM.__init__(self, channel)
        self.setExpiration(0.0)
        self.setSafetyEnabled(False)

    def stopMotor(self):
        """Stop the motor associated with this PWM object.
        This is called by the MotorSafety object when it has a timeout for
        this PWM and needs to stop it from running.
        """
        self.disable()

    def getDescription(self):
        return "PWM %d" % self.getChannel()

    def disable(self):
        self.setDisabled()



# validated: 2016-01-08 DS e15ca5a shared/java/edu/wpi/first/wpilibj/filters/Filter.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2015-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .interfaces.pidsource import PIDSource

__all__ = ['Filter']

class Filter:
    """Superclass for filters"""
    
    def __init__(self, source):
        """Constructor.
        
        :param source:
        :type source: :class:`.PIDSource`, callable
        """
        self.source = PIDSource.from_obj_or_callable(source)
        
    def setPIDSourceType(self, pidSourceType):
        self.source.setPIDSourceType(pidSourceType)
        
    def getPIDSourceType(self):
        return self.source.getPIDSourceType()
    
    def pidGet(self):
        raise NotImplementedError
    
    def get(self):
        """Returns the current filter estimate without also inserting new data as
        :meth:`pidGet` would do.
        
        :returns: The current filter estimate
        """
        raise NotImplementedError
    
    def reset(self):
        """Reset the filter state"""
        raise NotImplementedError
    
    def pidGetSource(self):
        """Calls PIDGet() of source
        
        :returns: Current value of source
        """
        return self.source.pidGet()
        


# notrack
# Placeholder for the CANJaguar class, which has been removed

__all__ = ["CANJaguar"]

class CANJaguar:

    def __init__(*args, **kwargs):
        raise ValueError("CANJaguar has been removed from wpilib.")




# validated: 2016-12-26 JW 0613f1d1827f athena/java/edu/wpi/first/wpilibj/Encoder.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings
import weakref

from .interfaces.counterbase import CounterBase
from .interfaces.pidsource import PIDSource
from .counter import Counter
from .digitalinput import DigitalInput
from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase

from ._impl.utils import match_arglist, HasAttribute

__all__ = ["Encoder"]

def _freeEncoder(encoder):
    hal.freeEncoder(encoder)

class Encoder(SensorBase):
    """Class to read quadrature encoders.
    
    Quadrature encoders are devices that count
    shaft rotation and can sense direction. The output of the Encoder class
    is an integer that can count either up or down, and can go negative for
    reverse direction counting. When creating Encoders, a direction can be
    supplied that inverts the sense of the output to make code more readable
    if the encoder is mounted such that forward movement generates negative
    values. Quadrature encoders have two digital outputs, an A Channel and a
    B Channel, that are out of phase with each other to allow the FPGA to do
    direction sensing.

    All encoders will immediately start counting - reset() them if you need
    them to be zeroed before use.

    Instance variables:
    
    - aSource: The A phase of the quad encoder
    - bSource: The B phase of the quad encoder
    - indexSource: The index source (available on some encoders)
    """

    class IndexingType:
        kResetWhileHigh = 0
        kResetWhileLow = 1
        kResetOnFallingEdge = 2
        kResetOnRisingEdge = 3

    EncodingType = CounterBase.EncodingType
    PIDSourceType = PIDSource.PIDSourceType

    def __init__(self, *args, **kwargs):
        """Encoder constructor. Construct a Encoder given a and b channels
        and optionally an index channel.

        The encoder will start counting immediately.

        The a, b, and optional index channel arguments may be either channel
        numbers or `DigitalSource` sources. There may also be a boolean
        reverseDirection, and an encodingType according to the following
        list.
        
        - aSource, bSource
        - aSource, bSource, reverseDirection
        - aSource, bSource, reverseDirection, encodingType
        - aSource, bSource, indexSource, reverseDirection
        - aSource, bSource, indexSource
        - aChannel, bChannel
        - aChannel, bChannel, reverseDirection
        - aChannel, bChannel, reverseDirection, encodingType
        - aChannel, bChannel, indexChannel, reverseDirection
        - aChannel, bChannel, indexChannel

        For positional arguments, if the passed object has a
        `getPortHandleForRouting` function, it is assumed to be a DigitalSource.

        Alternatively, sources and/or channels may be passed as keyword
        arguments.  The behavior of specifying both a source and a number
        for the same channel is undefined, as is passing both a positional
        and a keyword argument for the same channel.

        In addition, keyword parameters may be provided for reverseDirection
        and inputType.

        :param aSource: The source that should be used for the a channel.
        :param bSource: The source that should be used for the b channel.
        :param indexSource: The source that should be used for the index
            channel.
        :param aChannel: The digital input index that should be used for
            the a channel.
        :param bChannel: The digital input index that should be used for
            the b channel.
        :param indexChannel: The digital input index that should be used
            for the index channel.
        :param reverseDirection:
            Represents the orientation of the encoder and inverts the
            output values if necessary so forward represents positive
            values.  Defaults to False if unspecified.
        :param encodingType:
            Either k1X, k2X, or k4X to indicate 1X, 2X or 4X decoding. If
            4X is selected, then an encoder FPGA object is used and the
            returned counts will be 4x the encoder spec'd value since all
            rising and falling edges are counted. If 1X or 2X are selected
            then a counter object will be used and the returned value will
            either exactly match the spec'd count or be double (2x) the
            spec'd count.  Defaults to k4X if unspecified.
        :type encodingType: :class:`Encoder.EncodingType`
        """
        a_source_arg = ("aSource", HasAttribute("getPortHandleForRouting"))
        b_source_arg = ("bSource", HasAttribute("getPortHandleForRouting"))
        index_source_arg = ("indexSource", HasAttribute("getPortHandleForRouting"))
        a_channel_arg = ("aChannel", int)
        b_channel_arg = ("bChannel", int)
        index_channel_arg = ("indexChannel", int)

        argument_templates = [[a_source_arg, b_source_arg],
                              [a_source_arg, b_source_arg, ("reverseDirection", bool)],
                              [a_source_arg, b_source_arg, ("reverseDirection", bool), ("encodingType", int)],
                              [a_source_arg, b_source_arg, index_source_arg],
                              [a_source_arg, b_source_arg, index_source_arg, ("reverseDirection", bool)],
                              [a_channel_arg, b_channel_arg],
                              [a_channel_arg, b_channel_arg, ("reverseDirection", bool)],
                              [a_channel_arg, b_channel_arg, ("reverseDirection", bool), ("encodingType", int)],
                              [a_channel_arg, b_channel_arg, index_channel_arg],
                              [a_channel_arg, b_channel_arg, index_channel_arg, ("reverseDirection", bool)]]

        _, results = match_arglist('Encoder.__init__',
                                   args, kwargs, argument_templates)
        
        # keyword arguments
        aSource = results.pop("aSource", None)
        bSource = results.pop("bSource", None)
        indexSource = results.pop("indexSource", None)
        aChannel = results.pop("aChannel", None)
        bChannel = results.pop("bChannel", None)
        indexChannel = results.pop("indexChannel", None)
        reverseDirection = results.pop("reverseDirection", False)
        encodingType = results.pop("encodingType", self.EncodingType.k4X)

        # convert channels into sources
        self.allocatedA = False
        self.allocatedB = False
        self.allocatedIndex = False

        if aSource is None:
            if aChannel is None:
                raise ValueError("didn't specify A channel")
            aSource = DigitalInput(aChannel)
            self.allocatedA = True
        if bSource is None:
            if bChannel is None:
                raise ValueError("didn't specify B channel")
            bSource = DigitalInput(bChannel)
            self.allocatedB = True
        if indexSource is None and indexChannel is not None:
            indexSource = DigitalInput(indexChannel)
            self.allocatedIndex = True

        # save to instance variables
        self.aSource = aSource
        self.bSource = bSource
        self.indexSource = indexSource
        self.encodingType = encodingType
        self.distancePerPulse = 1.0 # distance of travel for each encoder tick
        self.pidSource = self.PIDSourceType.kDisplacement
        self._encoder = None
        self.counter = None
        self.index = 0

        if encodingType == self.EncodingType.k4X:
            self._encoder, self.index = hal.initializeEncoder(
                    aSource.getPortHandleForRouting(),
                    aSource.getAnalogTriggerTypeForRouting(),
                    bSource.getPortHandleForRouting(),
                    bSource.getAnalogTriggerTypeForRouting(),
                    reverseDirection, encodingType)
            self.__finalizer = \
                    weakref.finalize(self, _freeEncoder, self._encoder)
            self.setMaxPeriod(.5)
            self.encodingScale = 4
        elif encodingType in (self.EncodingType.k2X, self.EncodingType.k1X):
            # Use Counter object for 1x and 2x encoding
            self.counter = Counter(encodingType, aSource, bSource,
                                   reverseDirection)
            if encodingType == self.encodingType.k2X:
                self.encodingScale = 2
            else:
                self.encodingScale = 1
            self.index = self.counter.getFPGAIndex()
        else:
            raise ValueError("unrecognized encodingType: %s" % encodingType)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)
        
        if self.indexSource is not None:
            self.setIndexSource(self.indexSource)

        hal.report(hal.UsageReporting.kResourceType_Encoder,
                      self.index, encodingType)
        LiveWindow.addSensorChannel("Encoder", aSource.getChannel(),
                                    self)

    @property
    def encoder(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use encoder after free() has been called")
        return self._encoder

    def getFPGAIndex(self):
        """
        :returns: The Encoder's FPGA index
        """
        return hal.getEncoderFPGAIndex(self.encoder)

    def getEncodingScale(self):
        """
        :returns: The encoding scale factor 1x, 2x, or 4x, per the requested
            encodingType. Used to divide raw edge counts down to spec'd counts.
        """
        return hal.getEncoderEncodingScale(self.encoder)

    def free(self):
        LiveWindow.removeComponent(self)
        if self.aSource is not None and self.allocatedA:
            self.aSource.free()
            self.allocatedA = False
        if self.bSource is not None and self.allocatedB:
            self.bSource.free()
            self.allocatedB = False
        if self.indexSource is not None and self.allocatedIndex:
            self.indexSource.free()
            self.allocatedIndex = False
        self.aSource = None
        self.bSource = None
        self.indexSource = None
        self.__finalizer()
        self._encoder = None

    def getRaw(self):
        """Gets the raw value from the encoder. The raw value is the actual
        count unscaled by the 1x, 2x, or 4x scale factor.

        :returns: Current raw count from the encoder
        """
        return hal.getEncoderRaw(self.encoder)

    def get(self):
        """Gets the current count. Returns the current count on the Encoder.
        This method compensates for the decoding type.

        :returns: Current count from the Encoder adjusted for the 1x, 2x, or
            4x scale factor.
        """
        return hal.getEncoder(self.encoder)

    def reset(self):
        """Reset the Encoder distance to zero. Resets the current count to
        zero on the encoder.
        """
        hal.resetEncoder(self.encoder)

    def getPeriod(self):
        """Returns the period of the most recent pulse. Returns the period of
        the most recent Encoder pulse in seconds. This method compensates for
        the decoding type.

        .. deprecated::
            Use :func:`getRate` in favor of this method. This returns unscaled
            periods and :func:`getRate` scales using value from
            :func:`getDistancePerPulse`.

        :returns: Period in seconds of the most recent pulse.
        """
        warnings.warn("use getRate instead", DeprecationWarning)
        return hal.getEncoderPeriod(self.encoder)

    def setMaxPeriod(self, maxPeriod):
        """Sets the maximum period for stopped detection. Sets the value that
        represents the maximum period of the Encoder before it will assume
        that the attached device is stopped. This timeout allows users to
        determine if the wheels or other shaft has stopped rotating. This
        method compensates for the decoding type.

        :param maxPeriod: The maximum time between rising and falling edges
            before the FPGA will report the device stopped. This is expressed
            in seconds.
        """
        hal.setEncoderMaxPeriod(self.encoder, maxPeriod)

    def getStopped(self):
        """Determine if the encoder is stopped. Using the MaxPeriod value, a
        boolean is returned that is True if the encoder is considered stopped
        and False if it is still moving. A stopped encoder is one where the
        most recent pulse width exceeds the MaxPeriod.

        :returns: True if the encoder is considered stopped.
        """
        return hal.getEncoderStopped(self.encoder)

    def getDirection(self):
        """The last direction the encoder value changed.

        :returns: The last direction the encoder value changed.
        """
        return hal.getEncoderDirection(self.encoder)

    def getDistance(self):
        """Get the distance the robot has driven since the last reset.

        :returns: The distance driven since the last reset as scaled by the
            value from :func:`setDistancePerPulse`.
        """
        return hal.getEncoderDistance(self.encoder)

    def getRate(self):
        """Get the current rate of the encoder. Units are distance per second
        as scaled by the value from :func:`setDistancePerPulse`.

         :returns: The current rate of the encoder.
        """
        return hal.getEncoderRate(self.encoder)

    def setMinRate(self, minRate):
        """Set the minimum rate of the device before the hardware reports it
        stopped.

        :param minRate: The minimum rate. The units are in distance per
            second as scaled by the value from :func:`setDistancePerPulse`.
        """
        hal.setEncoderMinRate(self.encoder, minRate)

    def setDistancePerPulse(self, distancePerPulse):
        """Set the distance per pulse for this encoder. This sets the
        multiplier used to determine the distance driven based on the count
        value from the encoder. Do not include the decoding type in this
        scale. The library already compensates for the decoding type. Set
        this value based on the encoder's rated Pulses per Revolution and
        factor in gearing reductions following the encoder shaft. This
        distance can be in any units you like, linear or angular.

        :param distancePerPulse: The scale factor that will be used to convert
            pulses to useful units.
        """
        hal.setEncoderDistancePerPulse(self.encoder, distancePerPulse)

    def setReverseDirection(self, reverseDirection):
        """Set the direction sensing for this encoder. This sets the direction
        sensing on the encoder so that it could count in the correct software
        direction regardless of the mounting.

        :param reverseDirection: True if the encoder direction should be
            reversed
        """
        hal.setEncoderReverseDirection(self.encoder, reverseDirection)

    def setSamplesToAverage(self, samplesToAverage):
        """Set the Samples to Average which specifies the number of samples
        of the timer to average when calculating the period. Perform averaging
        to account for mechanical imperfections or as oversampling to increase
        resolution.

        TODO: Should this raise an exception, so that the user has to
        deal with giving an incorrect value?

        :param samplesToAverage: The number of samples to average from 1 to
            127.
        """
        hal.setEncoderSamplesToAverage(self.encoder, samplesToAverage)

    def getSamplesToAverage(self):
        """Get the Samples to Average which specifies the number of samples
        of the timer to average when calculating the period. Perform averaging
        to account for mechanical imperfections or as oversampling to increase
        resolution.

        :returns: The number of samples being averaged (from 1 to 127)
        """
        return hal.getEncoderSamplesToAverage(self.encoder)

    def setPIDSourceType(self, pidSource):
        """Set which parameter of the encoder you are using as a process
        control variable. The encoder class supports the rate and distance
        parameters.

        :param pidSource: An enum to select the parameter.
        """
        if pidSource not in (self.PIDSourceType.kDisplacement,
                             self.PIDSourceType.kRate):
            raise ValueError("Must be kRate or kDisplacement")
        self.pidSource = pidSource
        
    def getPIDSourceType(self):
        return self.pidSource

    def pidGet(self):
        """Implement the PIDSource interface.

         :returns: The current value of the selected source parameter.
        """
        if self.pidSource == self.PIDSourceType.kDisplacement:
            return self.getDistance()
        elif self.pidSource == self.PIDSourceType.kRate:
            return self.getRate()
        else:
            return 0.0

    def setIndexSource(self, source, indexing_type=IndexingType.kResetOnRisingEdge):
        """
        Set the index source for the encoder. When this source rises, the encoder count automatically resets.

        :param source: Either an initialized DigitalSource or a DIO channel number
        :type: Either a :class:`wpilib.DigitalInput` or number
        :param indexing_type: The state that will cause the encoder to reset
        :type: A value from :class:`wpilib.DigitalInput.IndexingType`
        """
        if hasattr(source, "getPortHandleForRouting"):
            self.indexSource = source
        else:
            self.indexSource = DigitalInput(source)

        hal.setEncoderIndexSource(self.encoder, self.indexSource.getPortHandleForRouting(),
                                  self.indexSource.getAnalogTriggerTypeForRouting(), indexing_type)

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        if self.encodingType == self.EncodingType.k4X:
            return "Quadrature Encoder"
        return "Encoder"

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    def getTable(self):
        return self.table

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Speed", self.getRate())
            table.putNumber("Distance", self.getDistance())
            table.putNumber("Distance per Tick", self.distancePerPulse)

    def startLiveWindowMode(self):
        pass

    def stopLiveWindowMode(self):
        pass



# novalidate
'''
    An implementation of java.util.TimerTask which is needed for
    the PID Controller
'''

import threading
from ..timer import Timer

import logging

class TimerTask(threading.Thread):
    
    def __init__(self, name, period, task_fn):
        super().__init__(name=name, daemon=True)
        
        if period <= 0:
            raise ValueError("Invalid period %s" % period)
        
        self.logger = logging.getLogger('wpilib.%s' % name)
        self.period = period
        self.task_fn = task_fn
        
        self.last_warning = -10
        
        self.stopped = False
    
    def cancel(self):
        self.stopped = True
        self.join()
    
    def run(self):
        
        self.logger.info("Timer task started")
        
        try:
            period = self.period
            wait_til = Timer.getFPGATimestamp() + period
            
            while not self.stopped:
                
                now = Timer.getFPGATimestamp()
                delay = wait_til - now
                
                if delay > 0:
                    Timer.delay(delay)
                else:
                    wait_til = now
                    # only emit warning once per second
                    if now - self.last_warning > 1:
                        self.last_warning = now
                        self.logger.warn("Delay time was negative (%02f: too much going on?)", delay)
                
                if self.stopped:
                    break
                
                self.task_fn()
                
                wait_til += period
        finally:
            self.logger.info("Timer task exited")



# validated: 2015-12-22 DS 6d854af shared/java/edu/wpi/first/wpilibj/interfaces/Potentiometer.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .pidsource import PIDSource

__all__ = ["Potentiometer"]

class Potentiometer(PIDSource):
    def get(self):
        raise NotImplementedError



# validated: 2016-12-01 AA 140c365e4b99 shared/java/edu/wpi/first/wpilibj/GenericHID.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["GenericHID"]

class GenericHID:
    """
    GenericHID Interface.
    """

    class RumbleType:
        """Represents a rumble output on the JoyStick."""

        #: Left Hand
        kLeftRumble = 0

        #: Right Hand
        kRightRumble = 1

    class HIDType:
        kUnknown = -1
        kXInputUnknown = 0
        kXInputGamepad = 1
        kXInputWheel = 2
        kXInputArcadeStick = 3
        kXInputFlightStick = 4
        kXInputDancePad = 5
        kXInputGuitar = 6
        kXInputGuitar2 = 7
        kXInputDrumKit = 8
        kXInputGuitar3 = 11
        kXInputArcadePad = 19
        kHIDJoystick = 20
        kHIDGamepad = 21
        kHIDDriving = 22
        kHIDFlight = 23
        kHID1stPerson = 24

        def __init__(self, value):
            self.value = value

    class Hand:
        """Which hand the Human Interface Device is associated with."""

        #: Left Hand
        kLeft = 0

        #: Right Hand
        kRight = 1

    def __init__(self, port):
        self.port = port

    def getX(self, hand=None):
        """Get the x position of HID.

        :param hand: which hand, left or right
        :returns: the x position
        """
        raise NotImplementedError

    def getY(self, hand=None):
        """Get the y position of the HID.

        :param hand: which hand, left or right
        :returns: the y position
        """
        raise NotImplementedError

    def getRawAxis(self, which):
        """Get the raw axis.

        :param which: index of the axis
        :returns: the raw value of the selected axis
        """
        raise NotImplementedError

    def getRawButton(self, button):
        """Is the given button pressed.

        :param button: which button number
        :returns: the angle of the POV in degrees, or -1 if the POV is not pressed.
        """
        raise NotImplementedError

    def getPOV(self, pov=0):
        """Get the angle in degrees of a POV on the HID.

        The POV angles start at 0 in the up direction, and increase clockwise (eg right is 90,
        upper-left is 315).

        :param pov: The index of the POV to read (starting at 0)
        :returns: the angle of the POV in degrees, or -1 if the POV is not pressed.
        """
        raise NotImplementedError

    def getPOVCount(self):
        """For the current HID, return the number of POVs."""
        pass

    def getPort(self):
        """Get the port number of the HID.

        :returns: The port number of the HID.
        """
        return self.port

    def getType(self):
        """Get the type of the HID.

        :returns: the type of the HID.
        """
        raise NotImplementedError

    def getName(self):
        """Get the name of the HID.

        :returns: the name of the HID.
        """
        raise NotImplementedError

    def setOutput(self, outputNumber, value):
        """Set a single HID output value for the HID.

        :param outputNumber: The index of the output to set (1-32)
        :param value: The value to set the output to
        """
        raise NotImplementedError

    def setOutputs(self, value):
        """Set all HID output values for the HID.

        :param value: The 32 bit output value (1 bit for each output)
        """
        raise NotImplementedError

    def setRumble(self, type, value):
        """Set the rumble output for the HID. The DS currently supports 2 rumble values, left rumble and
        right rumble.

        :param type: Which rumble value to set
        :param value: The normalized value (0 to 1) to set the rumble to
        """
        raise NotImplementedError



# validated: 2016-12-31 JW 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/AnalogOutput.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["AnalogOutput"]

def _freeAnalogOutput(port):
    hal.freeAnalogOutputPort(port)

class AnalogOutput(SensorBase):
    """Analog output"""

    channels = Resource(SensorBase.kAnalogOutputChannels)

    def __init__(self, channel):
        """Construct an analog output on a specified MXP channel.

        :param channel: The channel number to represent.
        """
        SensorBase.checkAnalogOutputChannel(channel)
        
        self.channel = channel

        port = hal.getPort(channel)
        self._port = hal.initializeAnalogOutputPort(port)

        LiveWindow.addSensorChannel("AnalogOutput", channel, self)
        hal.report(hal.UsageReporting.kResourceType_AnalogChannel,
                      channel, 1)
        
        self.__finalizer = weakref.finalize(self, _freeAnalogOutput, self._port)

    @property
    def port(self):
        if not self.__finalizer.alive:
            return None
        return self._port
    
    def free(self):
        """Channel destructor.
        """
        LiveWindow.removeComponent(self)
        if self.channel is None:
            return
        AnalogOutput.channels.free(self.channel)
        self.__finalizer()
        self.channel = None

    def getChannel(self):
        """Get the channel of this AnalogOutput.
        """
        return self.channel

    def setVoltage(self, voltage):
        hal.setAnalogOutput(self.port, voltage)

    def getVoltage(self):
        return hal.getAnalogOutput(self.port)

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Analog Output"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getVoltage())

    def startLiveWindowMode(self):
        # Analog Channels don't have to do anything special when entering the
        # LiveWindow.
        pass

    def stopLiveWindowMode(self):
        # Analog Channels don't have to do anything special when exiting the
        # LiveWindow.
        pass



# validated: 2016-12-27 JW aafca4ed7fff athena/java/edu/wpi/first/wpilibj/DigitalSource.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .resource import Resource
from .sensorbase import SensorBase
from .interruptablesensorbase import InterruptableSensorBase

__all__ = ["DigitalSource"]

def _freeDigitalSource(handle):
    hal.freeDIOPort(handle)

class DigitalSource(InterruptableSensorBase):
    """DigitalSource Interface. The DigitalSource represents all the possible
    inputs for a counter or a quadrature encoder. The source may be either a
    digital input or an analog input. If the caller just provides a channel,
    then a digital input will be constructed and freed when finished for the
    source. The source can either be a digital input or analog trigger but
    not both.
    
    .. not_implemented: initDigitalPort
    """

    channels = Resource(SensorBase.kDigitalChannels)

    def __init__(self, channel, input):
        """
            :param channel: Port for the digital input
            :type  channel: int
            :param input: True if input, False otherwise
            :type  input: int
        """
        super().__init__()

        SensorBase.checkDigitalChannel(channel)
        self.channel = channel
        self._handle = hal.initializeDIOPort(hal.getPort(channel), input)

        try:
            DigitalSource.channels.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Digital input %d is already allocated" % self.channel) from e

        self.__finalizer = weakref.finalize(self, _freeDigitalSource, self._handle)

    @property
    def handle(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use channel after free() has been called")
        return self._handle

    def free(self):
        DigitalSource.channels.free(self.channel)
        self.__finalizer()

    def isAnalogTrigger(self):
        raise NotImplementedError

    def getChannel(self):
        raise NotImplementedError




# validated: 2016-12-25 JW 963391cf3916 athena/java/edu/wpi/first/wpilibj/DoubleSolenoid.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase
from .solenoidbase import SolenoidBase

__all__ = ["DoubleSolenoid"]

class DoubleSolenoid(SolenoidBase):
    """Controls 2 channels of high voltage Digital Output on the PCM.

    The DoubleSolenoid class is typically used for pneumatics solenoids that
    have two positions controlled by two separate channels.
    """

    class Value:
        """Possible values for a DoubleSolenoid."""
        kOff = 0
        kForward = 1
        kReverse = 2

    def __init__(self, *args, **kwargs):
        """Constructor.

        Arguments can be supplied as positional or keyword.  Acceptable
        positional argument combinations are:
        
        - forwardChannel, reverseChannel
        - moduleNumber, forwardChannel, reverseChannel

        Alternatively, the above names can be used as keyword arguments.

        :param moduleNumber: The module number of the solenoid module to use.
        :param forwardChannel: The forward channel number on the module to control (0..7)
        :param reverseChannel: The reverse channel number on the module to control  (0..7)
        """
        # keyword arguments
        forwardChannel = kwargs.pop("forwardChannel", None)
        reverseChannel = kwargs.pop("reverseChannel", None)
        moduleNumber = kwargs.pop("moduleNumber", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 2:
            forwardChannel, reverseChannel = args
        elif len(args) == 3:
            moduleNumber, forwardChannel, reverseChannel = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        if moduleNumber is None:
            moduleNumber = SensorBase.getDefaultSolenoidModule()
        if forwardChannel is None:
            raise ValueError("must specify forward channel")
        if reverseChannel is None:
            raise ValueError("must specify reverse channel")

        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.checkSolenoidChannel(forwardChannel)
        SensorBase.checkSolenoidChannel(reverseChannel)

        super().__init__(moduleNumber)
        try:
            self.allocated.allocate(self, forwardChannel)
        except IndexError as e:
            raise IndexError("Solenoid channel %d on module %d is already allocated" % (forwardChannel, moduleNumber)) from e
        try:
            self.allocated.allocate(self, reverseChannel)
        except IndexError as e:
            raise IndexError("Solenoid channel %d on module %d is already allocated" % (reverseChannel, moduleNumber)) from e

        portHandle = hal.getPortWithModule(moduleNumber, forwardChannel)
        self.forwardHandle = hal.initializeSolenoidPort(portHandle)
        portHandle = hal.getPortWithModule(moduleNumber, reverseChannel)
        self.reverseHandle = hal.initializeSolenoidPort(portHandle)

        self.forwardChannel = forwardChannel
        self.reverseChannel = reverseChannel
        self.forwardMask = 1 << forwardChannel
        self.reverseMask = 1 << reverseChannel

        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        hal.report(hal.UsageReporting.kResourceType_Solenoid,
                      forwardChannel, moduleNumber)
        hal.report(hal.UsageReporting.kResourceType_Solenoid,
                      reverseChannel, moduleNumber)

        LiveWindow.addActuatorModuleChannel("DoubleSolenoid", moduleNumber,
                                            forwardChannel, self)

    def free(self):
        """Mark the solenoid as freed."""
        LiveWindow.removeComponent(self)
        self.allocated.free(self.forwardChannel)
        self.allocated.free(self.reverseChannel)
        hal.freeSolenoidPort(self.forwardHandle)
        hal.freeSolenoidPort(self.reverseHandle)
        super().free()

    def set(self, value):
        """Set the value of a solenoid.

        :param value: The value to set (Off, Forward, Reverse)
        :type  value: :class:`DoubleSolenoid.Value`
        """

        if value == self.Value.kOff:
            hal.setSolenoid(self.forwardHandle, False)
            hal.setSolenoid(self.reverseHandle, False)
        elif value == self.Value.kForward:
            hal.setSolenoid(self.reverseHandle, False)
            hal.setSolenoid(self.forwardHandle, True)
        elif value == self.Value.kReverse:
            hal.setSolenoid(self.forwardHandle, False)
            hal.setSolenoid(self.reverseHandle, True)
        else:
            raise ValueError("Invalid argument '%s'" % value)

    def get(self):
        """Read the current value of the solenoid.

        :returns: The current value of the solenoid.
        :rtype: :class:`DoubleSolenoid.Value`
        """
        if hal.getSolenoid(self.forwardHandle):
            return self.Value.kForward
        if hal.getSolenoid(self.reverseHandle):
            return self.Value.kReverse
        return self.Value.kOff

    def isFwdSolenoidBlackListed(self):
        """
        Check if the forward solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        blacklist = self.getPCMSolenoidBlackList()

        return (blacklist & self.forwardMask) != 0

    def isRevSolenoidBlackListed(self):
        """
        Check if the reverse solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        blacklist = self.getPCMSolenoidBlackList()

        return blacklist & (1 << self.reverseMask) != 0

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Double Solenoid"

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    def getTable(self):
        return self.table

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            #TODO: this is bad
            val = self.get()
            if val == self.Value.kForward:
                table.putString("Value", "F")
            elif val == self.Value.kReverse:
                table.putString("Value", "R")
            else:
                table.putString("Value", "O")

    def valueChanged(self, itable, key, value, bln):
        #TODO: this is bad also
        if value == "Reverse":
            self.set(self.Value.kReverse)
        elif value == "Forward":
            self.set(self.Value.kForward)
        else:
            self.set(self.Value.kOff)

    def startLiveWindowMode(self):
        self.set(self.Value.kOff) # Stop for safety
        super().startLiveWindowMode()

    def stopLiveWindowMode(self):
        super().stopLiveWindowMode()
        self.set(self.Value.kOff) # Stop for safety



# validated: 2016-11-26 DS 9e99df1cf70a athena/java/edu/wpi/first/wpilibj/PWMSpeedController.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .livewindow import LiveWindow
from .safepwm import SafePWM

__all__ = ["PWMSpeedController"]

class PWMSpeedController(SafePWM):
    """
        Common base class for all PWM Speed Controllers.
    """
    
    def __init__(self, channel):
        super().__init__(channel)
        self.isInverted = False
    
    def free(self):
        LiveWindow.removeComponent(self)
        super().free()

    def set(self, speed):
        """Set the PWM value.

        The PWM value is set using a range of -1.0 to 1.0, appropriately
        scaling the value for the FPGA.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        :type  speed: float
        """
        self.setSpeed(-speed if self.isInverted else speed)
        self.feed()

    def setInverted(self, isInverted):
        """
        Common interface for inverting the direction of a speed controller.

        :param isInverted: The state of inversion (True is inverted).
        """
        self.isInverted = isInverted

    def getInverted(self):
        """
        Common interface for inverting the direction of a speed controller.

        :returns: The state of inversion (True is inverted)
        """
        return self.isInverted

    def get(self):
        """Get the recently set value of the PWM.

        :returns: The most recently set value for the PWM between -1.0 and 1.0.
        :rtype: float
        """
        return self.getSpeed()

    def pidWrite(self, output):
        """Write out the PID value as seen in the PIDOutput base object.

        :param output: Write out the PWM value as was found in the
            :class:`PIDController`.
        :type  output: float
        """
        self.set(output)



# validated: 2015-12-22 DS 6d854af shared/java/edu/wpi/first/wpilibj/NamedSendable.java
from ..sendable import Sendable

__all__ = ["NamedSendable"]

class NamedSendable(Sendable):
    """The interface for sendable objects that gives the sendable a default
    name in the Smart Dashboard.
    """

    def getName(self):
        """
            :returns: The name of the subtable of SmartDashboard that the
                      :class:`.Sendable` object will use
        """
        raise NotImplementedError



# validated: 2015-12-24 DS 6d854af athena/java/edu/wpi/first/wpilibj/Resource.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import weakref

__all__ = ["Resource"]

class Resource:
    """Tracks resources in the program.
    
    The Resource class is a convenient way of keeping track of allocated
    arbitrary resources in the program. Resources are just indices that
    have an lower and upper bound that are tracked by this class. In the
    library they are used for tracking allocation of hardware channels
    but this is purely arbitrary. The resource class does not do any actual
    allocation, but simply tracks if a given index is currently in use.

    .. not_implemented: restartProgram
    """
    
    _resource_objects = []
    _global_resources = []
    
    @staticmethod
    def _reset():
        '''
            This clears all resources in the program and calls free() on any
            objects that have a free method.
        '''
        
        for resource in Resource._resource_objects:
            
            # free all the resources, if a free method is defined
            for ref in resource.numAllocated:
                if ref is None:
                    continue
                obj = ref()
                if obj is not None and hasattr(obj, 'free'):
                    obj.free()
            
            resource.numAllocated = [None]*len(resource.numAllocated)
        
        for ref in Resource._global_resources:
            obj = ref()
            if obj is not None and hasattr(obj, 'free'):
                obj.free()
        
    @staticmethod
    def _add_global_resource(obj):
        Resource._global_resources.append(weakref.ref(obj))
        

    def __init__(self, size):
        """Allocate storage for a new instance of Resource.
        Allocate a bool array of values that will get initialized to
        indicate that no resources have been allocated yet. The indicies
        of the resources are 0..size-1.

        :param size: The number of blocks to allocate
        """
        Resource._resource_objects.append(self)
        self.numAllocated = [None]*size

    def allocate(self, obj, index=None):
        """Allocate a resource.

        When index is None or unspecified, a free resource value within the
        range is located and returned after it is marked allocated.
        Otherwise, it is verified unallocated, then returned.

        :param obj: The object requesting the resource.
        :param index: The resource to allocate
        :returns: The index of the allocated block.
        :raises IndexError: If there are no resources available to be
            allocated or the specified index is already used.
        """
        if index is None:
            for i in range(len(self.numAllocated)):
                r = self.numAllocated[i]
                if r is None or r() is None:
                    self.numAllocated[i] = weakref.ref(obj)
                    return i
            raise IndexError("No available resources")

        if index >= len(self.numAllocated) or index < 0:
            raise IndexError("Index %d out of range" % index)
        r = self.numAllocated[index]
        if r is not None and r() is not None:
            raise IndexError("Resource at index %d already allocated" % index)
        self.numAllocated[index] = weakref.ref(obj)
        return index

    def free(self, index):
        """Force-free an allocated resource.
        After a resource is no longer needed, for example a destructor is
        called for a channel assignment class, free will release the resource
        value so it can be reused somewhere else in the program.

        :param index: The index of the resource to free.
        """
        self.numAllocated[index] = None



# validated: 2016-12-25 JW 963391cf3916 athena/java/edu/wpi/first/wpilibj/Solenoid.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref
import warnings

from .livewindow import LiveWindow
from .sensorbase import SensorBase
from .solenoidbase import SolenoidBase

__all__ = ["Solenoid"]

def _freeSolenoid(solenoidHandle):
    hal.freeSolenoidPort(solenoidHandle)

class Solenoid(SolenoidBase):
    """Solenoid class for running high voltage Digital Output.

    The Solenoid class is typically used for pneumatic solenoids, but could
    be used for any device within the current spec of the PCM.
    
    .. not_implemented: initSolenoid
    """

    def __init__(self, *args, **kwargs):
        """Constructor.

        Arguments can be supplied as positional or keyword.  Acceptable
        positional argument combinations are:
        
        - channel
        - moduleNumber, channel

        Alternatively, the above names can be used as keyword arguments.

        :param moduleNumber: The CAN ID of the PCM the solenoid is attached to
        :type moduleNumber: int
        :param channel: The channel on the PCM to control (0..7)
        :type channel: int
        """
        # keyword arguments
        channel = kwargs.pop("channel", None)
        moduleNumber = kwargs.pop("moduleNumber", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 1:
            channel = args[0]
        elif len(args) == 2:
            moduleNumber, channel = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        if moduleNumber is None:
            moduleNumber = SensorBase.getDefaultSolenoidModule()
        if channel is None:
            raise ValueError("must specify channel")

        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.checkSolenoidChannel(channel)

        super().__init__(moduleNumber)
        self.channel = channel

        try:
            self.allocated.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Solenoid channel %d on module %d is already allocated" % (channel, moduleNumber)) from e

        portHandle= hal.getPortWithModule(moduleNumber, channel)
        self._solenoidHandle = hal.initializeSolenoidPort(portHandle)

        LiveWindow.addActuatorModuleChannel("Solenoid", moduleNumber, channel,
                                            self)
        hal.report(hal.UsageReporting.kResourceType_Solenoid, channel,
                   moduleNumber)
        
        self.__finalizer = weakref.finalize(self, _freeSolenoid, self._solenoidHandle)
        
    @property
    def solenoidHandle(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use channel after free() has been called")
        return self._solenoidHandle

    def free(self):
        """Mark the solenoid as freed."""
        LiveWindow.removeComponent(self)
        self.allocated.free(self.channel)
        
        self.__finalizer()
        self._solenoidHandle = None
        
        super().free()

    def set(self, on):
        """Set the value of a solenoid.

        :param on: True will turn the solenoid output on. False will turn the solenoid output off.
        :type on: bool
        """
        hal.setSolenoid(self.solenoidHandle, on)

    def get(self):
        """Read the current value of the solenoid.

        :returns: True if the solenoid output is on or false if the solenoid output is off.
        :rtype: bool
        """
        return hal.getSolenoid(self.solenoidHandle)

    def isBlackListed(self):
        """
        Check if the solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        value = self.getPCMSolenoidBlackList() & (1 << self.channel)
        return value != 0

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Solenoid"

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    def getTable(self):
        return self.table

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putBoolean("Value", self.get())

    def valueChanged(self, itable, key, value, bln):
        self.set(True if value else False)

    def startLiveWindowMode(self):
        self.set(False) # Stop for safety
        super().startLiveWindowMode()

    def stopLiveWindowMode(self):
        super().stopLiveWindowMode()
        self.set(False) # Stop for safety



# validated: 2016-12-22 JW aafca4ed7fff athena/java/edu/wpi/first/wpilibj/InterruptableSensorBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .sensorbase import SensorBase

__all__ = ["InterruptableSensorBase"]

class InterruptableSensorBase(SensorBase):
    """Base for sensors to be used with interrupts"""

    def __init__(self):
        """Create a new InterrupatableSensorBase"""
        # The interrupt resource
        self._interrupt = None
        self._interrupt_finalizer = None
        # Flags if the interrupt being allocated is synchronous
        self.isSynchronousInterrupt = False

    def getAnalogTriggerTypeForRouting(self):
        raise NotImplementedError

    def getPortHandleForRouting(self):
        raise NotImplementedError

    @property
    def interrupt(self):
        if self._interrupt_finalizer is None:
            return None
        if not self._interrupt_finalizer.alive:
            return None
        return self._interrupt

    def requestInterrupts(self, handler=None):
        """Request one of the 8 interrupts asynchronously on this digital
        input.

        :param handler: (optional)
            The function that will be called whenever there is an interrupt
            on this device.  Request interrupts in synchronous mode where the
            user program interrupt handler will be called when an interrupt
            occurs. The default is interrupt on rising edges only.  If not
            specified, the user program will have to explicitly wait for the
            interrupt to occur using waitForInterrupt.
        """
        if self.interrupt is not None:
            raise ValueError("The interrupt has already been allocated")

        self.allocateInterrupts(handler is not None)

        assert self.interrupt is not None

        hal.requestInterrupts(self.interrupt, self.getPortHandleForRouting(),
                              self.getAnalogTriggerTypeForRouting())
        self.setUpSourceEdge(True, False)
        if handler is not None:
            hal.attachInterruptHandler(self.interrupt, handler)

    def allocateInterrupts(self, watcher):
        """Allocate the interrupt

        :param watcher: True if the interrupt should be in synchronous mode
            where the user program will have to explicitly wait for the interrupt
            to occur.
        """
        if self.interrupt is not None:
            raise ValueError("The interrupt has already been allocated")

        self.isSynchronousInterrupt = watcher
        self._interrupt = hal.initializeInterrupts(watcher)
        self._interrupt_finalizer = weakref.finalize(self, hal.cleanInterrupts,
                                                     self._interrupt)

    def cancelInterrupts(self):
        """Cancel interrupts on this device. This deallocates all the
        chipobject structures and disables any interrupts.
        """
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        self._interrupt_finalizer()
        self.interrupt = None

    def waitForInterrupt(self, timeout, ignorePrevious=True):
        """In synchronous mode, wait for the defined interrupt to occur.
        You should **NOT** attempt to read the sensor from another thread
        while waiting for an interrupt. This is not threadsafe, and can cause 
        memory corruption

        :param timeout: Timeout in seconds
        :param ignorePrevious: If True (default), ignore interrupts that
            happened before waitForInterrupt was called.
        """
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        return hal.waitForInterrupt(self.interrupt, timeout, ignorePrevious)

    def enableInterrupts(self):
        """Enable interrupts to occur on this input. Interrupts are disabled
        when the RequestInterrupt call is made. This gives time to do the
        setup of the other options before starting to field interrupts.
        """
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        if self.isSynchronousInterrupt:
            raise ValueError("You do not need to enable synchronous interrupts")
        hal.enableInterrupts(self.interrupt)

    def disableInterrupts(self):
        """Disable Interrupts without without deallocating structures."""
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        if self.isSynchronousInterrupt:
            raise ValueError("You can not disable synchronous interrupts")
        hal.disableInterrupts(self.interrupt)

    def readRisingTimestamp(self):
        """Return the timestamp for the rising interrupt that occurred most
        recently.  This is in the same time domain as getClock().  The
        rising-edge interrupt should be enabled with setUpSourceEdge.

        :returns: Timestamp in seconds since boot.
        """
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        return hal.readInterruptRisingTimestamp(self.interrupt)

    def readFallingTimestamp(self):
        """Return the timestamp for the falling interrupt that occurred most
        recently.  This is in the same time domain as getClock().  The
        falling-edge interrupt should be enabled with setUpSourceEdge.

        :returns: Timestamp in seconds since boot.
        """
        if self.interrupt is None:
            raise ValueError("The interrupt is not allocated.")
        return hal.readInterruptFallingTimestamp(self.interrupt)

    def setUpSourceEdge(self, risingEdge, fallingEdge):
        """Set which edge to trigger interrupts on

        :param risingEdge: True to interrupt on rising edge
        :param fallingEdge: True to interrupt on falling edge
        """
        if self.interrupt is not None:
            hal.setInterruptUpSourceEdge(self.interrupt,
                                         1 if risingEdge else 0,
                                         1 if fallingEdge else 0)
        else:
            raise ValueError("You must call RequestInterrupts before setUpSourceEdge")



# validated: 2016-11-20 KC b78f580d476f shared/java/edu/wpi/first/wpilibj/command/PrintCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .instantcommand import InstantCommand

__all__ = ["PrintCommand"]

class PrintCommand(InstantCommand):
    """A PrintCommand is a command which prints out a string when it is
    initialized, and then immediately finishes.

    It is useful if you want a :class:`.CommandGroup` to print out a string when it
    reaches a certain point.
    """

    def __init__(self, message):
        """Instantiates a PrintCommand which will print the given message when
        it is run.
        
        :param message: the message to print
        """
        super().__init__('Print("%s")' % message)
        self.message = message

    def initialize(self):
        print(self.message)



# notrack
# Placeholder for the CANTalon class, which has moved to its own package

__all__ = ["CANTalon"]

class CANTalon:

    def __init__(*args, **kwargs):
        raise ValueError("CANTalon has moved to ctre.CANTalon, you must install robotpy-ctre")




#----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .controller import Controller

__all__ = ['PIDInterface']

class PIDInterface(Controller):
    
    def setPID(self, p, i, d):
        raise NotImplementedError
    
    def getP(self):
        raise NotImplementedError
    
    def getI(self):
        raise NotImplementedError
    
    def getD(self):
        raise NotImplementedError
    
    def setSetpoint(self, setpoint):
        raise NotImplementedError
    
    def getSetpoint(self):
        raise NotImplementedError
    
    def getError(self):
        raise NotImplementedError
    
    def enable(self):
        raise NotImplementedError
        
    def disable(self):
        raise NotImplementedError
        
    def isEnabled(self):
        raise NotImplementedError
        
    def reset(self):
        raise NotImplementedError



# validated: 2016-12-31 JW 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/Joystick.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import math

from .interfaces.joystickbase import JoystickBase

__all__ = ["Joystick"]

class Joystick(JoystickBase):
    """Handle input from standard Joysticks connected to the Driver Station.
    
    This class handles standard input that comes from the Driver Station. Each
    time a value is requested the most recent value is returned. There is a
    single class instance for each joystick and the mapping of ports to
    hardware buttons depends on the code in the Driver Station.
    """

    kDefaultXAxis = 0
    kDefaultYAxis = 1
    kDefaultZAxis = 2
    kDefaultTwistAxis = 2
    kDefaultThrottleAxis = 3
    kDefaultTriggerButton = 1
    kDefaultTopButton = 2

    class AxisType:
        """Represents an analog axis on a joystick."""
        kX = 0
        kY = 1
        kZ = 2
        kTwist = 3
        kThrottle = 4
        kNumAxis = 5

    class ButtonType:
        """Represents a digital button on the Joystick"""
        kTrigger = 0
        kTop = 1
        kNumButton = 2

    def __init__(self, port, numAxisTypes=None, numButtonTypes=None):
        """Construct an instance of a joystick.

        The joystick index is the USB port on the Driver Station.

        This constructor is intended for use by subclasses to configure the
        number of constants for axes and buttons.

        :param port: The port on the Driver Station that the joystick is
            plugged into.
        :type  port: int
        :param numAxisTypes: The number of axis types.
        :type  numAxisTypes: int
        :param numButtonTypes: The number of button types.
        :type  numButtonTypes: int
        """
        super().__init__(port)
        from .driverstation import DriverStation
        self.ds = DriverStation.getInstance()

        if numAxisTypes is None:
            self.axes = [0]*self.AxisType.kNumAxis
            self.axes[self.AxisType.kX] = self.kDefaultXAxis
            self.axes[self.AxisType.kY] = self.kDefaultYAxis
            self.axes[self.AxisType.kZ] = self.kDefaultZAxis
            self.axes[self.AxisType.kTwist] = self.kDefaultTwistAxis
            self.axes[self.AxisType.kThrottle] = self.kDefaultThrottleAxis
        else:
            self.axes = [0]*numAxisTypes

        if numButtonTypes is None:
            self.buttons = [0]*self.ButtonType.kNumButton
            self.buttons[self.ButtonType.kTrigger] = self.kDefaultTriggerButton
            self.buttons[self.ButtonType.kTop] = self.kDefaultTopButton
        else:
            self.buttons = [0]*numButtonTypes

        self.outputs = 0
        self.leftRumble = 0
        self.rightRumble = 0

        hal.report(hal.UsageReporting.kResourceType_Joystick, port)

    def getX(self, hand=None):
        """Get the X value of the joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :param hand: Unused
        :returns: The X value of the joystick.
        :rtype: float
        """
        return self.getRawAxis(self.axes[self.AxisType.kX])

    def getY(self, hand=None):
        """Get the Y value of the joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :param hand: Unused
        :returns: The Y value of the joystick.
        :rtype: float
        """
        return self.getRawAxis(self.axes[self.AxisType.kY])

    def getZ(self, hand=None):
        return self.getRawAxis(self.axes[self.AxisType.kZ])

    def getTwist(self):
        """Get the twist value of the current joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :returns: The Twist value of the joystick.
        :rtype: float
        """
        return self.getRawAxis(self.axes[self.AxisType.kTwist])

    def getThrottle(self):
        """Get the throttle value of the current joystick.

        This depends on the mapping of the joystick connected to the current
        port.

        :returns: The Throttle value of the joystick.
        :rtype: float
        """
        return self.getRawAxis(self.axes[self.AxisType.kThrottle])

    def getRawAxis(self, axis):
        """Get the value of the axis.

        :param axis: The axis to read, starting at 0.
        :type  axis: int
        :returns: The value of the axis.
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, axis)

    def getAxis(self, axis):
        """For the current joystick, return the axis determined by the
        argument.

        This is for cases where the joystick axis is returned programmatically,
        otherwise one of the previous functions would be preferable (for
        example :func:`getX`).

        :param axis: The axis to read.
        :type axis: :class:`Joystick.AxisType`
        :returns: The value of the axis.
        :rtype: float
        """
        if axis == self.AxisType.kX:
            return self.getX()
        elif axis == self.AxisType.kY:
            return self.getY()
        elif axis == self.AxisType.kZ:
            return self.getZ()
        elif axis == self.AxisType.kTwist:
            return self.getTwist()
        elif axis == self.AxisType.kThrottle:
            return self.getThrottle()
        else:
            raise ValueError("Invalid axis specified! Must be one of wpilib.Joystick.AxisType, or use getRawAxis instead")
        
    def getAxisCount(self):
        """For the current joystick, return the number of axis"""
        return self.ds.getStickAxisCount(self.getPort())

    def getTrigger(self, hand=None):
        """Read the state of the trigger on the joystick.

        Look up which button has been assigned to the trigger and read its
        state.

        :param hand: This parameter is ignored for the Joystick class and is
            only here to complete the GenericHID interface.
        :returns: The state of the trigger.
        :rtype: bool
        """
        return self.getRawButton(self.buttons[self.ButtonType.kTrigger])

    def getTop(self, hand=None):
        """Read the state of the top button on the joystick.

        Look up which button has been assigned to the top and read its state.

        :param hand: This parameter is ignored for the Joystick class and is
            only here to complete the GenericHID interface.
        :returns: The state of the top button.
        :rtype: bool
        """
        return self.getRawButton(self.buttons[self.ButtonType.kTop])

    def getPOV(self, pov=0):
        return self.ds.getStickPOV(self.getPort(), pov)

    def getPOVCount(self):
        return self.ds.getStickPOVCount(self.getPort())

    def getBumper(self, hand=None):
        """This is not supported for the Joystick.

        This method is only here to complete the GenericHID interface.

        :param hand: This parameter is ignored for the Joystick class and is
            only here to complete the GenericHID interface.
        :returns: The state of the bumper (always False)
        :rtype: bool
        """
        return False

    def getRawButton(self, button):
        """Get the button value (starting at button 1).

        The buttons are returned in a single 16 bit value with one bit
        representing the state of each button. The appropriate button is
        returned as a boolean value.

        :param button: The button number to be read (starting at 1).
        :type  button: int
        :returns: The state of the button.
        :rtype: bool
        """
        return self.ds.getStickButton(self.getPort(), button)
    
    def getButtonCount(self):
        """For the current joystick, return the number of buttons
        
        :rtype int
        """
        return self.ds.getStickButtonCount(self.getPort())

    def getButton(self, button):
        """Get buttons based on an enumerated type.

        The button type will be looked up in the list of buttons and then read.

        :param button: The type of button to read.
        :type  button: :class:`.Joystick.ButtonType`
        :returns: The state of the button.
        :rtype: bool
        """
        if button == self.ButtonType.kTrigger:
            return self.getTrigger()
        elif button == self.ButtonType.kTop:
            return self.getTop()
        else:
            raise ValueError("Invalid button specified! Must be one of wpilib.Joystick.ButtonType, or use getRawButton instead")

    def getMagnitude(self):
        """Get the magnitude of the direction vector formed by the joystick's
        current position relative to its origin.

        :returns: The magnitude of the direction vector
        :rtype: float
        """
        return math.sqrt(math.pow(self.getX(), 2) + math.pow(self.getY(), 2))

    def getDirectionRadians(self):
        """Get the direction of the vector formed by the joystick and its
        origin in radians.

        :returns: The direction of the vector in radians
        :rtype: float
        """
        return math.atan2(self.getX(), -self.getY())

    def getDirectionDegrees(self):
        """Get the direction of the vector formed by the joystick and its
        origin in degrees.

        :returns: The direction of the vector in degrees
        :rtype: float
        """
        return math.degrees(self.getDirectionRadians())

    def getAxisChannel(self, axis):
        """Get the channel currently associated with the specified axis.

        :param axis: The axis to look up the channel for.
        :type  axis: int
        :returns: The channel for the axis.
        :rtype: int
        """
        return self.axes[axis]

    def setAxisChannel(self, axis, channel):
        """Set the channel associated with a specified axis.

        :param axis: The axis to set the channel for.
        :type  axis: int
        :param channel: The channel to set the axis to.
        :type  channel: int
        """
        self.axes[axis] = channel

    def getIsXbox(self):
        """
        Get the value of isXbox for the current joystick.

        :returns: A boolean that is true if the controller is an xbox controller.
        """
        return self.ds.getJoystickIsXbox(self.getPort())

    def getAxisType(self, axis):
        """Get the axis type of a joystick axis.

        :returns: the axis type of a joystick axis.
        """
        return self.ds.getJoystickAxisType(self.getPort(), axis)

    def getType(self):
        """Get the type of the HID.

        :returns: the type of the HID.
        """
        return self.ds.getJoystickType(self.getPort())

    def getName(self):
        """
        Get the name of the HID.

        :returns: The name of the HID.
        """
        return self.ds.getJoystickName(self.getPort())


    def setOutput(self, outputNumber, value):
        self.outputs = (self.outputs & ~(value << (outputNumber-1))) | (value << (outputNumber-1))
        self.flush_outputs()

    def setOutputs(self, value):
        self.outputs = value
        self.flush_outputs()

    def setRumble(self, type, value):
        """Set the rumble output for the joystick. The DS currently supports 2 rumble values,
        left rumble and right rumble

        :param type: Which rumble value to set
        :type  type: :class:`.Joystick.RumbleType`
        :param value: The normalized value (0 to 1) to set the rumble to
        :type  value: float
        """
        if value < 0:
            value = 0
        elif value > 1:
            value = 1
        if type == self.RumbleType.kLeftRumble:
            self.leftRumble = int(value*65535)
        elif type == self.RumbleType.kRightRumble:
            self.rightRumble = int(value*65535)
        else:
            raise ValueError("Invalid wpilib.Joystick.RumbleType: {}".format(type))
        self.flush_outputs()

    def flush_outputs(self):
        """Flush all joystick HID & rumble output values to the HAL"""
        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)



# validated: 2017-01-02 DV c3160bad4463 shared/java/edu/wpi/first/wpilibj/smartdashboard/SendableChooser.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .sendable import Sendable

__all__ = ["SendableChooser"]

class SendableChooser(Sendable):
    """A useful tool for presenting a selection of options to be displayed on
    the SmartDashboard

    For instance, you may wish to be able to select between multiple
    autonomous modes. You can do this by putting every possible Command
    you want to run as an autonomous into a SendableChooser and then put
    it into the SmartDashboard to have a list of options appear on the
    laptop. Once autonomous starts, simply ask the SendableChooser what
    the selected value is.

    Example::

        # This shows the user two options on the SmartDashboard
        chooser = wpilib.SendableChooser()
        chooser.addObject('option1', '1')
        chooser.addObject('option2', '2')

        wpilib.SmartDashboard.putData('Choice', chooser)

        # .. later, ask to see what the user selected?
        value = chooser.getSelected()

    """

    # The key for the default value
    DEFAULT = "default"
    # The key for the selected option
    SELECTED = "selected"
    # The key for the option array
    OPTIONS = "options"
    # A table linking strings to the objects the represent

    def __init__(self):
        """Instantiates a SendableChooser.
        """
        self.map = {}
        self.defaultChoice = None

    def addObject(self, name, object):
        """Adds the given object to the list of options. On the
        SmartDashboard on the desktop, the object will appear as the
        given name.

        :param name: the name of the option
        :param object: the option
        """
        # if we don't have a default, set the default automatically
        if self.defaultChoice is None:
            self.addDefault(name, object)
            return
        self.map[name] = object

        table = self.getTable()
        if table is not None:
            table.putStringArray(self.OPTIONS, self.map.keys())

    def addDefault(self, name, object):
        """Add the given object to the list of options and marks it as the
        default.  Functionally, this is very close to addObject(...) except
        that it will use this as the default option if none other is
        explicitly selected.

        :param name: the name of the option
        :param object: the option
        """
        if name is None:
            raise ValueError("Name cannot be None")
        self.defaultChoice = name
        table = self.getTable()
        if table is not None:
            table.putString(self.DEFAULT, self.defaultChoice)
        self.addObject(name, object)

    def getSelected(self):
        """Returns the object associated with the selected option. If there
        is none selected, it will return the default. If there is none
        selected and no default, then it will return None.

        :returns: the object associated with the selected option
        """
        table = self.getTable()
        if table is None:
            return self.map[self.defaultChoice]
        selected = table.getString(self.SELECTED, None)
        return self.map.get(selected, self.map[self.defaultChoice])

    def getSmartDashboardType(self):
        return "String Chooser"

    def initTable(self, table):
        self.table = table
        if table is not None:
            table.putStringArray(self.OPTIONS, self.map.keys())
            if self.defaultChoice is not None:
                table.putString(self.DEFAULT, self.defaultChoice)



# validated: 2016-12-25 JW e44a6e227a89 athena/java/edu/wpi/first/wpilibj/DigitalGlitchFilter.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2015. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.
#----------------------------------------------------------------------------

import hal
import threading

from .digitalsource import DigitalSource
from .encoder import Encoder
from .counter import Counter
from .sensorbase import SensorBase

__all__ = ['DigitalGlitchFilter']

class DigitalGlitchFilter(SensorBase):
    '''
    Class to enable glitch filtering on a set of digital inputs.
    This class will manage adding and removing digital inputs from a FPGA glitch
    filter. The filter lets the user configure the time that an input must remain
    high or low before it is classified as high or low.
    '''
    
    mutex = threading.Lock()
    filterAllocated = [False]*3
    
    def __init__(self):
        self.channelIndex = -1
        with self.mutex:
            for i, v in enumerate(self.filterAllocated):
                if not v:
                    self.channelIndex = i
                    self.filterAllocated[i] = True
                    hal.report(hal.UsageReporting.kResourceType_DigitalFilter,
                               self.channelIndex, 0)
                    break
            else:
                raise ValueError("No more filters available")
    
    def free(self):
        if self.channelIndex >= 0:
            with self.mutex:
                self.filterAllocated[self.channelIndex] = False
            
            self.channelIndex = -1
    
    @staticmethod
    def _setFilter(input, channelIndex):
        if input is not None: # Counter might have just one input
            # analog triggers are not supported for DigitalGlitchFilter
            if input.isAnalogTrigger():
                raise ValueError("Analog Triggers are not supported for DigitalGlitchFilters")
            hal.setFilterSelect(input.getPortHandleForRouting(), channelIndex)
            
            selected = hal.getFilterSelect(input.getPortHandleForRouting())
            if selected != channelIndex:
                raise ValueError('setFilterSelect(%s) failed -> %s' % (channelIndex, selected))
    
    def add(self, input):
        '''
        Assigns the :class:`.DigitalSource`, :class:`.Encoder`, or
        :class:`.Counter` to this glitch filter.
        
        :param input: The object to add
        '''
        if isinstance(input, DigitalSource):
            DigitalGlitchFilter._setFilter(input, self.channelIndex + 1)
        elif isinstance(input, Encoder):
            DigitalGlitchFilter._setFilter(input.aSource, self.channelIndex + 1)
            DigitalGlitchFilter._setFilter(input.bSource, self.channelIndex + 1)
        elif isinstance(input, Counter):
            DigitalGlitchFilter._setFilter(input.upSource, self.channelIndex + 1)
            DigitalGlitchFilter._setFilter(input.downSource, self.channelIndex + 1)
        else:
            raise ValueError("Cannot add %s to glitch filter" % input)
        
    def remove(self, input):
        '''
        Removes this filter from the given input object
        '''
        
        if isinstance(input, DigitalSource):
            DigitalGlitchFilter._setFilter(input, 0)
        elif isinstance(input, Encoder):
            DigitalGlitchFilter._setFilter(input.aSource, 0)
            DigitalGlitchFilter._setFilter(input.bSource, 0)
        elif isinstance(input, Counter):
            DigitalGlitchFilter._setFilter(input.upSource, 0)
            DigitalGlitchFilter._setFilter(input.downSource, 0)
        else:
            raise ValueError("Cannot remove %s from glitch filter" % input)
        
    def setPeriodCycles(self, fpga_cycles):
        '''
        Sets the number of FPGA cycles that the input must hold steady to pass
        through this glitch filter.
        
        :param fpga_cycles: The number of FPGA cycles.
        '''
        hal.setFilterPeriod(self.channelIndex, fpga_cycles)
        
    def setPeriodNanoSeconds(self, nanoseconds):
        '''
        Sets the number of nanoseconds that the input must hold steady to pass
        through this glitch filter.
        
        :param nanoseconds: The number of nanoseconds.
        '''
        fpga_cycles = int(nanoseconds * self.kSystemClockTicksPerMicrosecond / 4 / 1000)
        self.setPeriodCycles(fpga_cycles)
    
    def getPeriodCycles(self):
        '''
        Gets the number of FPGA cycles that the input must hold steady to pass
        through this glitch filter.
       
        :returns: The number of cycles.
        '''
        return hal.getFilterPeriod(self.channelIndex)
    
    def getPeriodNanoSeconds(self):
        '''
        Gets the number of nanoseconds that the input must hold steady to pass
        through this glitch filter.
       
        :returns: The number of nanoseconds.
        '''
        fpga_cycles = self.getPeriodCycles()
        return fpga_cycles * 1000 / (self.kSystemClockTicksPerMicrosecond / 4)



# validated: 2016-12-25 JW 69422dc0636c athena/java/edu/wpi/first/wpilibj/Compressor.java
import hal

from .sensorbase import SensorBase

__all__ = ["Compressor"]

class Compressor(SensorBase):
    """Class for operating a compressor connected to a PCM (Pneumatic Control Module).
    
    The PCM will automatically run in closed loop mode by default whenever a
    Solenoid object is created. For most cases the Compressor object does not
    need to be instantiated or used in a robot program. This class is only required 
    in cases where the robot program needs a more detailed status of the compressor or to
    enable/disable closed loop control.

    Note: you cannot operate the compressor directly from this class as doing
    so would circumvent the safety provided by using the pressure switch and closed loop control.
    You can only turn off closed loop control, thereby stopping the compressor from operating.
    """

    def __init__(self, module=None):
        """Makes a new instance of the compressor using the provided CAN device ID.
        
        :param module: The PCM CAN device ID. (0 - 62 inclusive)
        """
        self.table = None
        if module is None:
            module = SensorBase.getDefaultSolenoidModule()
        self.compressorHandle = hal.initializeCompressor(module)
        self.module = module

    def start(self):
        """Start the compressor running in closed loop control mode.

        Use the method in cases where you would like to manually stop and
        start the compressor for applications such as conserving battery
        or making sure that the compressor motor doesn't start during
        critical operations.
        """
        self.setClosedLoopControl(True)

    def stop(self):
        """Stop the compressor from running in closed loop control mode.

        Use the method in cases where you would like to manually stop and
        start the compressor for applications such as conserving battery
        or making sure that the compressor motor doesn't start during
        critical operations.
        """
        self.setClosedLoopControl(False)

    def enabled(self):
        """Get the enabled status of the compressor.
        
        :returns: True if the compressor is on
        :rtype: bool
        """
        return hal.getCompressor(self.compressorHandle)

    def getPressureSwitchValue(self):
        """ Get the pressure switch value.
 
        :returns: True if the pressure is low
        :rtype: bool
        """
        return hal.getCompressorPressureSwitch(self.compressorHandle)

    def getCompressorCurrent(self):
        """Get the current being used by the compressor.
        
        :returns: Current consumed by the compressor in amps
        :rtype: float
        """
        return hal.getCompressorCurrent(self.compressorHandle)

    def setClosedLoopControl(self, on):
        """Set the PCM in closed loop control mode.
        
        :param on: If True sets the compressor to be in closed loop control
                   mode (default)
        :type  on: bool
        """
        hal.setCompressorClosedLoopControl(self.compressorHandle, True if on else False)

    def getClosedLoopControl(self):
        """Gets the current operating mode of the PCM.
        
        :returns: True if compressor is operating on closed-loop mode
        :rtype: bool
        """
        return hal.getCompressorClosedLoopControl(self.compressorHandle)

    def getCompressorCurrentTooHighFault(self):
        """
        :returns: True if PCM is in fault state : Compressor Drive is
            disabled due to compressor current being too high
        """
        return hal.getCompressorCurrentTooHighFault(self.compressorHandle)

    def getCompressorCurrentTooHighStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor is
            disabled due to compressor current being too high
        """
        return hal.getCompressorCurrentTooHighStickyFault(self.compressorHandle)

    def getCompressorShortedFault(self):
        """
        :returns: True if PCM is in fault state : Compressor output
            appears to be shorted
        """
        return hal.getCompressorShortedFault(self.compressorHandle)

    def getCompressorShortedStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor output
            appears to be shorted
        """
        return hal.getCompressorShortedStickyFault(self.compressorHandle)

    def getCompressorNotConnectedFault(self):
        """
        :returns: True if PCM is in fault state : Compressor does not appear
            to be wired, i.e. compressor is not drawing enough current.
        """
        return hal.getCompressorNotConnectedFault(self.compressorHandle)

    def getCompressorNotConnectedStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor does not appear
            to be wired, i.e. compressor is not drawing enough current.
        """
        return hal.getCompressorNotConnectedStickyFault(self.compressorHandle)

    def clearAllPCMStickyFaults(self):
        """Clear ALL sticky faults inside PCM that Compressor is wired to.

        If a sticky fault is set, then it will be persistently cleared. The compressor might 
        momentarily disable while the flags are being cleared. Doo not call this method too
        frequently, otherwise normal compressor functionality may be prevented.

        If no sticky faults are set then this call will have no effect.
        """
        hal.clearAllPCMStickyFaults(self.module)

    def getSmartDashboardType(self):
        return "Compressor"

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    def getTable(self):
        return self.table

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putBoolean("Enabled", self.enabled())
            table.putBoolean("Pressure Switch", self.getPressureSwitchValue())
